#!/usr/bin/env python
# -*- coding:utf-8 -*-

from lcyframe import route
from lcyframe import funts
from base import BaseHandler, helper


@route("/permission/groups")
class PermissionGroupsHandler(BaseHandler):

    @helper.admin(1)
    def get(self):
        """
        角色详情

        Request extra query params:
        - group_id # type: int required:True 角色id

        :return:
        :rtype:
        """

        group_mp = self.model.GroupsModel.get_groups_mapping(self.params["content_type"])
        if not group_mp or str(self.params["group_id"]) not in group_mp:
            raise self.api_error.ErrorCommonPermission
        group = group_mp[str(self.params["group_id"])]
        group = self.model.GroupsModel.get_members_count(None, group)[0]
        self.write_success(group)

    @helper.admin(1)
    def post(self):
        """
        新增角色

        Request extra query params:
        - name # type: str required:True 权限名称
        - content_type # type: int required:True 0系统端角色，1业务的角色
        - describes # type: str required:False 角色说明

        :return:
        :rtype:
        """
        name = self.params.pop("name")
        content_type = self.params.pop("content_type")
        try:
            self.model.GroupsModel.create_groups(name, content_type, **self.params)
        except Exception as e:
            raise self.api_error.ErrorCommonPermission(str(e))
        self.write_success()

    @helper.admin(1)
    def put(self):
        """
        修改角色

        Request extra query params:
        - group_id # type: int required:True id
        - name # type: str required:False 权限名称
        - state # type: int required:False 1启用，0禁用
        - describes # type: str required:False 角色说明

        :return:
        :rtype:
        """
        group_id = self.params.pop("group_id")
        group = self.model.GroupsModel.get_data_by_kv(id=group_id)
        if not group:
            raise self.api_error.ErrorNotData
        content_type = group["content_type"]
        try:
            self.model.GroupsModel.update_group(group_id, content_type, **self.params)
        except Exception as e:
            raise self.api_error.ErrorCommonPermission(str(e))
        self.write_success()

    @helper.admin(1)
    def delete(self):
        """
        删除角色

        Request extra query params:
        - group_id # type: int required:True id

        :return:
        :rtype:
        """

        # self.model.GroupsModel.delete(self.params["oid"])
        self.write_success()


@route("/permission/groups/list")
class PermissionGroupsListHandler(BaseHandler):

    @helper.admin(1)
    def get(self):
        """
        角色列表

        Request extra query params:
        - page # type: int required:False 翻页码
        - count # type: int required:False 每页显示条数
        - state # type: int required:False 状态
        - content_type # type: int required:True 0系统端角色，1业务的角色
        - company_id # type: int required:True 指定商户，默认所有角色。

        :return:
        :rtype:
        """

        count = self.params.get("count", 10)
        page = self.params.get("page", 1)

        sql_and = {}
        orderby = {
            "id": -1
        }
        if self.params.get("content_type"):
            sql_and["content_type"] = self.params["content_type"]
        if self.params.get("state"):
            sql_and["state"] = self.params["state"]
        if self.params.get("company_id"):
            sql_and["company_id"] = self.params["company_id"]

        datas, counts = self.model.GroupsModel.get_group_list_by_page(
            page, count,
            sql_and=sql_and,
            orderby=orderby)

        datas = self.model.GroupsModel.get_members_count(self.params.get("company_id"), *datas)

        self.write_pagination(datas, counts)


@route("/permission/groups/members")
class PermissionGroupsMembersHandler(BaseHandler):

    @helper.admin(1)
    def get(self):
        """
        角色的成员
        查看指定角色或指定商户的成员列表
        :return:
        :rtype:
        """
        company_id = self.params.pop("company_id", None)
        group = self.model.GroupsModel.get_data_by_kv(id=self.params["group_id"])
        datas = self.model.GroupsModel.get_members(company_id, group)
        company_mp = self.model.CompanyModel.get_batch_by_in("name", id=[data["company_id"] for data in datas])
        for data in datas:
            data["company_name"] = company_mp[data["company_id"]]["name"]
        self.write_success({"datas": datas})

    @helper.admin(1)
    def post(self):
        """
        该角色分配成员
        """
        user_ids = list(map(int, self.params["user_id"].split(",")))
        company_id = self.params.pop("company_id")
        group_id = self.params.pop("group_id")
        group = self.model.GroupsModel.get_data_by_kv(id=group_id)
        if not group:
            raise self.api_error.ErrorCommonPermission("角色不存在")

        user_mp = self.model.AdminModel.get_batch_by_in("id", "group_id", "company_id", id=user_ids)
        for id, user in user_mp.items():
            if user["company_id"] == 1 and group["content_type"] == 1:
                raise self.api_error.ErrorCommonPermission("该角色只适用业务端")
            if user["company_id"] > 1 and group["content_type"] == 0:
                raise self.api_error.ErrorCommonPermission("该角色只适用系统端")
            if user["company_id"] != company_id:
                raise self.api_error.ErrorCommonPermission("用户%s不属于该商户" % user["username"])
            # 支持多个角色
            # group_ids = set(user["group_id"].split(","))
            # group_ids.add(str(group_id))
            # group_ids = group_ids-set([0, "0"])
            # self.model.AdminModel.update_data(user["id"], group_id=",".join(list(group_ids)))
            self.model.AdminModel.update_data(user["id"], group_id=str(group_id))
        self.write_success()


@route("/permission/groups/check")
class PermissionGroupsCheckHandler(BaseHandler):

    @helper.admin(1)
    def post(self):
        """
        编辑角色权限
        权限勾选或取消,修改成功后，建议刷新本地权限缓存

        Request extra query params:
        - group_id # type: int required:True 角色id
        - key # type: str required:True 接口标识符
        - method # type: str required:True 接口方法
        - value # type: int required:True 1勾选，0取消

        :return:
        :rtype:
        """

        group_id = self.params["group_id"]
        key = self.params["key"]
        method = self.params["method"]
        value = self.params["value"]

        self.model.GroupsModel.update_group_permission(group_id, key, method, value)
        self.write_success()


@route("/permission/groups/checks")
class PermissionGroupsCheckSHandler(BaseHandler):

    @helper.admin(1)
    def post(self):
        """
        批量修改权限
        编辑角色权限
        权限勾选或取消,修改成功后，建议刷新本地权限缓存

        Request extra query params:
        - group_id # type: int required:True 角色id

        :return:
        :rtype:
        """

        group_id = self.params["group_id"]
        permissions = self.params["permissions"]
        self.model.GroupsModel.update_group_permissions(group_id, permissions)
        self.write_success()


@route("/permission/groups/company")
class PermissionGroupsCompanyHandler(BaseHandler):

    @helper.admin(1)
    def get(self):
        """
        商户的所有角色
        查看某商户的所有角色，不分页

        Request extra query params:
        - company_id # type: int required:True 商户id

        :return:
        :rtype:
        """
        content_type = self.params["content_type"]
        if content_type:
            if not self.params["company_id"]:
                raise self.api_error.ErrorCommonPermission("需提供商户id")
            company_groups_mp = self.model.CompanyGroupsModel.get_company_groups(content_type)
            groups = company_groups_mp.get(str(self.params["company_id"]), [])
        else:
            groups = self.model.CompanyGroupsModel.get_system_groups()

        groups_mp = self.model.GroupsModel.get_groups_mapping(content_type)

        enable_groups = []
        for index, group_id in enumerate(groups):
            if str(group_id) in groups_mp:
                # 已被禁用的，也显示
                g = groups_mp[str(group_id)]
                # if g["state"] != 1:
                #     continue
                enable_groups.append(g)

        enable_groups = self.model.GroupsModel.get_members_count(self.params["company_id"], *enable_groups)

        self.write_success(data={"datas": enable_groups})

    @helper.admin(1)
    def post(self):
        """
        把角色分配给商户

        Request extra query params:
        - group_id # type: str required:True 角色id， 多个角色用英文逗号隔开
        - company_id # type: int required:True 商户id

        :return:
        :rtype:
        """
        group_id = self.params.pop("group_id", "").split(",")
        company_id = self.params.pop("company_id")
        if company_id == 1:
            raise self.api_error.ErrorCommonPermission("禁止修改系统端角色")
        self.model.CompanyGroupsModel.update_company_group(company_id, group_id)
        self.write_success()

    @helper.admin(1)
    def delete(self):
        """
        移除角色
        将角色从商户下移除，属于该角色的用户将失去该角色的所有权限

        Request extra query params:
        - group_id # type: int required:True 角色id
        - company_id # type: int required:True 商户id

        :return:
        :rtype:
        """

        group_id = self.params.pop("group_id", "").split(",")
        company_id = self.params.pop("company_id")
        self.model.CompanyGroupsModel.remove_company_group(company_id, group_id)
        self.write_success()


@route("/permission")
class PermissionHandler(BaseHandler):

    @helper.admin(1)
    def get(self):
        """
        权限详情

        Request extra query params:
        - permission_id # type: int required:True id

        :return:
        :rtype:
        """

        data = self.model.RBACPermissionModel.get_data_by_kv(id=self.params["permission_id"])

        self.write_success(data)

    @helper.admin(1)
    def post(self):
        """
        新增

        Request extra query params:
        - content_type # type: int required:True 0系统端角色，1业务的角色
        - key # type: str required:True 接口标识符
        - name # type: str required:True 名称
        - permission_class # type: int required:True 类别，tag标签，api接口权限
        - method # type: str required:False 接口方法， 当permission_class=api时必须提供
        - parent_id # type: str required:False 父级id
        - describes # type: str required:False 角色说明
        - state # type: int required:False 角色说明
        - hook_func # type: str required:False 钩子方法

        :return:
        :rtype:
        """
        key = self.params.pop("key")
        name = self.params.pop("name")
        permission_class = self.params.pop("permission_class")
        content_type = self.params.pop("content_type")

        if content_type not in [0, 1]:
            raise self.api_error.ErrorCommonPermission("参数错误")
        if self.model.RBACPermissionModel.get_datas_by_spec(sql_and=dict(key=key)):
            raise self.api_error.ErrorCommonPermission("`%s`已被使用" % key)
        if self.model.RBACPermissionModel.get_datas_by_spec(sql_and=dict(name=name, content_type=content_type)):
            raise self.api_error.ErrorCommonPermission("`%s`已被使用" % name)
        if self.params.get("hook_func") and self.model.RBACPermissionModel.get_datas_by_spec(
                sql_and=dict(hook_func=self.params.get("hook_func"))):
            raise self.api_error.ErrorCommonPermission("`%s`已存在" % self.params.get("hook_func"))
        if permission_class not in ["tag", "api", "logic逻辑权限"]:
            raise self.api_error.ErrorCommonPermission("permission_class只能是tag，api，logic中的一个")
        if permission_class != "tag":
            if not self.params.get("method"):
                raise self.api_error.ErrorCommonPermission("没有提供方法名method")
            for method in self.params.get("method").split(","):
                if method not in ["get", "delete", "put", "post"]:
                    raise self.api_error.ErrorCommonPermission("不支持该方法名" % method)

        self.model.RBACPermissionModel.create_permission(key, name, permission_class, content_type, **self.params)
        self.write_success()

    @helper.admin(1)
    def put(self):
        """
        修改
        修改权限名称，描述，状态，钩子

        Request extra query params:
        - content_type # type: int required:True 0系统端角色，1业务的角色
        - key # type: str required:False 接口标识符
        - name # type: str required:False 名称
        - permission_class # type: int required:False 类别，tag标签，api接口权限
        - method # type: str required:False 接口方法， 当permission_class=api时必须提供
        - parent_id # type: int required:False 父级id
        - describes # type: str required:False 角色说明
        - state # type: int required:False 角色说明
        - hook_func # type: str required:False 钩子方法

        :return:
        :rtype:
        """

        permission_id = self.params.pop("permission_id")
        content_type = self.params.pop("content_type")
        permission = self.model.RBACPermissionModel.get_data_by_kv(id=permission_id)
        if not permission:
            raise self.api_error.ErrorCommonPermission
        if self.params.get("parent_id") and self.params["parent_id"] == permission_id:
            raise self.api_error.ErrorCommonPermission("父级id错误")

        # 提交0，代表不设置父级id
        if "parent_id" in self.params and self.params["parent_id"] == 0:
            self.params["parent_id"] = None
        self.model.RBACPermissionModel.update_permission(permission_id, content_type, **self.params)
        self.write_success()

    @helper.admin(1)
    def delete(self):
        """
        删除

        Request extra query params:
        - permission_id # type: int required:True id

        :return:
        :rtype:
        """
        permission_id = self.params.pop("permission_id")
        content_type = self.params.pop("content_type")
        self.model.RBACPermissionModel.remove_permission(permission_id, content_type)
        self.write_success()


@route("/permission/list")
class PermissionListHandler(BaseHandler):

    @helper.admin(1)
    def get(self):
        """
        权限条目清单

        Request extra query params:
        - state # type: int required:False 状态
        - content_type # type: int required:True 0系统端权限条目，1业务端权限条目

        :return:
        :rtype:
        """
        permissions = []
        for item in self.model.RBACPermissionModel.get_permissions_list(self.params["content_type"]):
            if item["state"] == 1:
                permissions.append(item)

        self.write_success({"datas": permissions})


@route("/permission/tags/list")
class PermissionTagsListHandler(BaseHandler):

    @helper.admin(1)
    def get(self):
        """
        查看权限列表

        Request extra query params:
        - content_type # type: int required:True 0系统端权限条目，1业务端权限条目

        :return:
        :rtype:
        """
        page, count = self.params.get("page", 1), self.params.get("count", 10)
        sql_and = {
            "permission_class": self.params["permission_class"]
        }
        orderby = {
            "id": -1
        }
        if self.params.get("name"):
            sql_and["name"] = {"REGEXP": self.params.get("name")}
        if self.params.get("content_type") != None:
            sql_and["content_type"] = self.params.get("content_type")

        datas, counts = self.model.RBACPermissionModel.get_list_by_page(
            page, count,
            sql_and=sql_and,
            orderby=orderby)

        permissions_mp = self.model.RBACPermissionModel.get_batch_by_in("id", "name",
                                                                        id=[d["parent_id"] for d in datas if
                                                                            d["parent_id"]])
        for d in datas:
            if d["parent_id"] in permissions_mp:
                d["parent_name"] = permissions_mp[d["parent_id"]]["name"]

        self.write_pagination(datas, counts)


@route("/permission/tags")
class PermissionTagsHandler(BaseHandler):

    @helper.admin(1)
    def get(self):
        """
        查看权限类别(下拉)

        Request extra query params:
        - content_type # type: int required:True 0系统端权限条目，1业务端权限条目

        :return:
        :rtype:
        """
        permissions_tags = []
        datas = self.model.RBACPermissionModel.get_permissions_list(self.params["content_type"])

        def col_tags(permissions):
            for item in permissions:
                if item["permission_class"] == "tag":
                    permissions_tags.append(item)

                if item.get("children"):
                    col_tags(item["children"])

        col_tags(datas)

        self.write_success({"datas": permissions_tags})


@route("/permission/enable")
class PermissionEnableHandler(BaseHandler):

    @helper.admin(1)
    def get(self):
        """
        是否启用权限系统

        :return:
        :rtype:
        """
        data = self.model.RBACPermissionModel.is_enable(self.params["content_type"])
        self.write_success(data={
            "enable": 1 if data else 0
        })

    @helper.admin(1)
    def post(self):
        """
        启用停用
        停用权限系统后，所有访问都不做拦截

        Request extra query params:
        - state # type: int required:True 状态， 1启用，0停用

        :return:
        :rtype:
        """
        state = self.params.pop("state")
        if state:
            self.model.RBACPermissionModel.enable_permission(self.params["content_type"])
        else:
            self.model.RBACPermissionModel.disable_permission(self.params["content_type"])
        self.write_success()


@route("/permission/loads")
class PermissionLoadseHandler(BaseHandler):

    @helper.admin(1)
    def post(self):
        """
        收集并导入当前所有handler的权限条目
        当后端新增新的api逻辑后，调用该方法可自动载入新增的权限条目。该功能仅支持系统端
        Request extra query params:
        - state # type: int required:True 状态， 1启用，0停用

        :return:
        :rtype:
        """
        new_permission = {}

        routes = {}
        routes_hander_mp = {}
        exists_permission = {}
        for item in self.application.handlers:
            key = item.name
            if key in ["DocsHandler", "BaseHandler"]:
                continue
            routes_hander_mp[key] = item.handler_class
            funcs = vars(item.target).keys()
            routes.setdefault(key, [])
            for method in ["get", "post", "put", "delete"]:
                if method in funcs:
                    routes[key].append(method)
                else:
                    continue

        # 已存在的权限
        permissions = self.model.RBACPermissionModel.get_datas_by_spec(sql_and={"content_type": 0})
        for permission in permissions:
            exists_permission[permission["key"]] = permission["method"]

        # 筛选出未导入的选线条目
        for key, methods in routes.items():
            if not methods:
                continue
            if key not in exists_permission:
                new_permission[key] = methods
            else:
                for method in exists_permission[key].split(","):
                    if key in new_permission and method in new_permission[key]:
                        new_permission[key].remove(method)

        datas = self.model.RBACPermissionModel.autoload_permission(routes_hander_mp, new_permission)
        self.write_success({"datas": datas})


@route("/permission/export_and_import")
class PermissionExportImportHandler(BaseHandler):

    @helper.admin(1)
    def get(self):
        """
        导出权限条目
        Request extra query params:
        - state # type: int required:True 状态， 1启用，0停用

        :return:
        :rtype:
        """
        permissions = self.model.RBACPermissionModel.export_permission(self.params["content_type"])
        self.write_success({"datas": permissions})

    @helper.admin(1)
    def post(self):
        """
        上传并导入权限
        file：文件导入，文件格式必须为json形式数组
        permissions：参数提交，json形式数组
        注意：已存在的key不会被导入
        """
        is_delete = self.params.get("is_delete", False)
        content_type = self.params["content_type"]
        permissions = self.params.get("permissions", None)
        file = self.params.get("file", None)
        permissions = permissions or self.utils.to_python(file["body"].decode())
        try:
            data = self.model.RBACPermissionModel.import_permissions(permissions, content_type, is_delete)
        except Exception as e:
            raise self.api_error.ErrorCommonPermission(str(e))

        self.write_success("权限导入完成，成功导入:%s条" % (len(data)))


import json

import redis

from scanner.configs import Scaner_Config, out
from scheduler.common import keys

"""
收集最终结果
"""


class RedisCon(object):
    """单例"""

    _redis = None

    @classmethod
    def get_connection(cls):
        if cls._redis is None:
            cls._redis = redis.Redis(
                host=Scaner_Config.REDIS_HOST,
                port=Scaner_Config.REDIS_PORT,
                db=7,
                password=Scaner_Config.REDIS_PASSWORD,
            )
        return cls._redis


class CacheData(object):
    redis = RedisCon.get_connection()

    @classmethod
    def add_domain(cls, task_id, domain):
        """
        域名记录
        """
        return cls.redis.setex(keys.scan_task_domain % task_id, Scaner_Config.EXPIRE, domain)

    @classmethod
    def add_subdomains(cls, task_id, *datas):
        """
        子域名记录
        Returns:
            subdomains: [
                {
                    "subdomainname": "sub.domain.com",  # 必须，其他的中间逻辑依赖该值
                    "ip": "10.10.10.10",
                    "city": "南宁",
                    "country": "中国",
                    "province": "广西",
                    "ips": "电信",
                    "cdn": "True",
                    "domain_id": 10
                },
                ...
            ]
        """
        key = keys.scan_task_subdomains % task_id
        set_res = cls.redis.sadd(key, *[json.dumps(d) for d in datas])
        cls.redis.expire(key, Scaner_Config.EXPIRE)
        return set_res

    @classmethod
    def add_ports(cls, task_id, *datas):
        """
        端口记录
        Returns:
            datas: [
                {
                    "ip": "10.0.0.1",
                    "port": 80,
                    "service": "service",
                    "protocol": "http",
                    "version": "2.0",
                    "subdomainname": "sub.domain.com",
                    "subdomain_id": 20,     # 若为空，则业务内需自行查库；来源：sudomain.id,相同ip可能存在多条不同的子域，取最后一条
                    "company_id": 1,
                    "domain_id": 0,
                    "task_id": 31
                }
            ]
        """
        key = keys.scan_task_ports % task_id
        set_res = cls.redis.sadd(key, *[json.dumps(d) for d in datas])
        cls.redis.expire(key, Scaner_Config.EXPIRE)
        return set_res

    @classmethod
    def get_domain(cls, task_id):
        key = keys.scan_task_domain % task_id
        dms = (cls.redis.get(key) or b"").decode()
        # cls.redis.expire(key, 3600)
        return dms

    @classmethod
    def get_subdomains(cls, task_id):
        key = keys.scan_task_subdomains % task_id
        sdm_list = [json.loads(d) for d in cls.redis.smembers(key)]
        # cls.redis.expire(key, 3600)
        return sdm_list

    @classmethod
    def get_ports(cls, task_id):
        key = keys.scan_task_ports % task_id
        pts = [json.loads(d) for d in cls.redis.smembers(key)]
        # cls.redis.expire(key, 3600)
        return pts

    @classmethod
    def get_result(cls, task_id):
        """
        搜集本次任务的结果
        Args:
            task_id:
        Returns:
        """
        domain = cls.get_domain(task_id)
        subdomains = cls.get_subdomains(task_id)
        ports = cls.get_ports(task_id)
        return {"domain": domain, "subdomains": subdomains, "ports": ports}

    @classmethod
    def __del__(cls):
        # cls.redis.delete('task_domain*')
        # cls.redis.delete('task_ports*')
        # cls.redis.delete('task_subdomains*')
        pass


if __name__ == "__main__":
    cache = CacheData()
    tk_id = "task_id"
    # cache.add_domain(task_id, "domain")
    # print(cache.get_domain(task_id))
    #
    # cache.add_subdomains(task_id, *[
    #             {
    #                 "subdomainname": "sub.domain.com",
    #                 "ip": "10.10.10.10",
    #                 "city": "南宁",
    #                 "country": "中国",
    #                 "province": "广西",
    #                 "ips": "电信",
    #                 "cdn": "True",
    #                 "domain_id": 10
    #             }
    #         ])
    # print(cache.get_subdomains(task_id))
    #
    # cache.add_ports(task_id, *[dict(
    #                 port=80,
    #                 service="service",
    #                 protocol="http",
    #                 version='2.0',
    #                 subdomain_id=20,
    #                 company_id=1,
    #                 domain_id=10,
    #                 task_id=31
    #             )])
    # print(cache.get_ports(task_id))

    out(cache.get_result(tk_id))



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import random, string
from test_script import *
from lcyframe.libs import cprint, utils


def post_demo(*args, **kwargs):
    """
    添加:测试post
    """
    headers = {}
    files = {}
    # 上传文件，参数放在body内，以multipart/form-data方式提交;excel=self.params["excel"]
    # files["excel"] = open("../请指定文件路径")
    
    params = {
        "str": "",     # string,选填,描述
        "str": "",     # string,选填,手机号
        "int": 0,     # integer,必填,整形
        "float": 0,     # float,必填,浮点型
        "json": "",     # json,必填,json格式
        "form-data": "",     # json,选填,位置：body_json，参数放在body内，以multipart/form-data方式，json格式。json类型的参数，设为非必须时，需提供默认值
        "form-data2": "",     # string,选填,位置：body_json，参数放在body内，以multipart/form-data方式，json格式。json类型的参数，设为非必须时，需提供默认值
        "www-form": "",     # string,选填,位置：body_form，参数放在body表单内，以multipart/x-www-form-urlencoded方式提交。手机号正则限定
        }
    return send(methed="post", url="/demo", params=params, headers=headers, files=files)

def get_demo(*args, **kwargs):
    """
    查看:测试get
    """
    headers = {"token": 1}
    files = {}
    params = {
        "_id": "",     # str,选填,id
        "a": 1,     # integer,必填,角色id
        "b": "",     # string,选填,供应商id
        "d": 1,     # int,选填,城市全拼列表
        }
    return send(methed="get", url="/demo", params=params, headers=headers, files=files)


def get_demo_list(*args, **kwargs):
    """
    列表:
    """
    headers = {}
    files = {}
    params = {
        "page": 0,     # integer,选填,翻页码
        "count": 0,     # integer,选填,每页显示条数
        "search": "",     # str,选填,搜索关键字
        "search": "",     # str,选填,搜索关键字
        "state": 0,     # int,选填,申请状态
        "time_range": "",     # str,选填,申请时间范围，逗号隔开。2020-12-12 12:12:12,2022-12-12 12:12:12
        }
    return send(methed="get", url="/demo/list", params=params, headers=headers, files=files)


def batch_groups():
    """
    测试簇
    一键测试
    """
    post_demo()       # 添加
    get_demo()       # 查看
    get_demo_list()       # 列表
    

if __name__ == "__main__":
    # 添加
    # post_demo()
    # 查看
    get_demo()
    # 列表
    # get_demo_list()
    


#!/usr/bin/env python
# -*- coding:utf-8 -*-
import time, datetime
import jwt

from jwt.exceptions import (
    DecodeError,
    ExpiredSignatureError,
    ImmatureSignatureError,
    InvalidAudienceError,
    InvalidIssuedAtError,
    InvalidIssuerError,
    MissingRequiredClaimError,
)
HS256 = "HS256"
default_headers = {"alg": HS256}

class JwtToken(object):

    def __init__(self, secret, exp=3600, algorithm=None, headers=None):
        self.secret = secret    # key
        self.exp = exp      # 小于等于0时，不过期
        self.algorithm = algorithm or HS256
        self.headers = headers or default_headers

    def encode(self, payload):
        return jwt.encode(self._get_payload(payload), self.secret, self.algorithm, self.headers)

    def decode(self, token):
        try:
            payloay = jwt.decode(token, self.secret, self.algorithm)
            return payloay
        except ExpiredSignatureError as e:
            return False
        except Exception as e:
            return {}

    def is_expire(self, token):
        """
        是否过期
        """
        if self.decode(token) is False:
            return True
        else:
            return False

    def get_expire(self, token):
        """
        获取到期时间戳
        """
        payload = self.decode(token)
        if payload:
            return payload["exp"]

    def is_validate(self, token):
        """
        是否合法
        """
        if self.decode(token):
            return True
        else:
            return False

    def refresh_token(self, token):
        """
        刷新token，仅支持带有效期的token
        """
        payload = self.decode(token)
        if not payload:
            raise Exception("token无效或已过期")
        if not payload.get("exp"):
            raise Exception("该token是永久的，无需刷新")
        return self.encode(payload)

    def _get_payload(self, payload):
        if not isinstance(payload, dict):
            payload = {"data": payload}
        if "exp" in payload:
            exp = payload["exp"]
            if isinstance(exp, int):
                if exp:
                    payload.update({"exp": int(time.time()) + exp})
                else:
                    payload.pop("exp")
        elif self.exp:
            payload.update({"exp": int(time.time()) + self.exp})
        return payload

if __name__ == '__main__':
    import json, datetime
    t = JwtToken("f465ca164f0e", 3600)
    # token = t.encode({"uid": 54887111})
    # 指定有效期为12小时
    # token = t.encode({"uid": 54887111, "exp": 3600*12})
    # 指定2022-09-06 08:00:00到期
    token = t.encode({"uid": 54887111, "exp": datetime.datetime.fromordinal(datetime.datetime.today().toordinal()) + datetime.timedelta(days=7)})
    print(token)
    print(t.decode(token))
    print(t.is_expire(token))
    print(t.is_validate(token))
    print(t.refresh_token(token))
    print(t.get_expire(token))



import os
import json
import re
import sys
import datetime
import random
import django
project_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(base_dir)
os.environ.setdefault("DJANGO_SETTINGS_MODULE", f"settings.dev")
django.setup()

from django.conf import settings
from django.db import connection
from django.db.models import Q
from django.forms.models import model_to_dict
from permission.permission_struct import service_permissions, system_permissions, service_groups, system_groups
from permission.models import Permissions, Groups, CompanyGroups

keys = []
names = []

def check_field(permissions):
    """
    校验权限合法性
    :param item:
    :return:
    """
    for p in permissions:
        if p.get("children", []):
            check_field(p["children"])
        else:
            if not p["name"] or not p["key"] or not p["permission_class"]:
                raise Exception("关键字都必填")
            # if p["permission_class"] == "tag" and not p["children"]:
            #     raise Exception("标签下尚未设置权限")
            if p["permission_class"] == "api" and not p["method"]:
                raise Exception("API级别权限必须指定方法：get、post、put、delete")
            if p["permission_class"] == "logic" and not p["method"]:
                raise Exception("暂不支持逻辑权限")

            if p["permission_class"] == "api":
                unique_name = p["key"] + ":" + p["method"]
            else:
                unique_name = p["key"]
            if unique_name not in keys:
                keys.append(unique_name)
            else:
                raise Exception("存在重复的key：%s" % p["key"])

            if p["name"] not in keys:
                names.append(p["name"])
            else:
                raise Exception("存在重复的name：%s" % p["name"])
    return True

def import_permissions(content_type=1):
    """
    权限json导入db
    "content_type": 1 or 0              # 1 业务端使用，0系统管理端
    :return:
    """
    if Permissions.objects.filter(content_type=content_type).first():
        raise Exception("请勿重复导入")

    if content_type == 1:
        permissions = service_permissions
    else:
        permissions = system_permissions
    assert check_field(permissions)
    def loop_create_item(item, parent=None):
        parent = Permissions.objects.create(
            name=item["name"],
            key=item["key"],
            method=item.get("method", ""),
            # parent=item.get("parent", ""),    # 去父级的key+method
            parent_id=parent["id"] if parent else None,
            permission_class=item.get("permission_class", ""),
            hook_func=item.get("hook_func", ""),
            state=item.get("state", 1),
            desc=item.get("desc", ""),
            content_type=content_type
        )
        print("写入权限:%s" % item["name"])
        for children in item.get("children", []):
            loop_create_item(children, model_to_dict(parent))

    for p in permissions:
        loop_create_item(p)

    print("权限导入完成，总计:%s条" % len(keys))

def import_groups(content_type=1):
    """
    创建初始化角色
    :return:
    """
    if content_type == 1:
        groups = service_groups
    else:
        groups = system_groups
    for group in groups:
        g = Groups.objects.create(**group)
        print("角色写入：", model_to_dict(g))

def set_company_groups(company_id, content_type=1):
    """
    给商户分配角色
    :param company_id:
    :return:
    """
    groups = []
    for g in Groups.objects.filter(state=1, content_type=content_type):
        groups.append(model_to_dict(g))
    for i in random.sample(groups, len(groups)-1):
        i["company_id"] = company_id
        g = CompanyGroups.objects.create(**{"company_id": company_id, "group_id": i["id"]})
        print("给商户分配角色:", model_to_dict(g))

def set_groups_permission(content_type=1):
    """
    设置角色权限：随机
    :return:
    """
    permissions = Permissions.objects.filter(content_type=content_type, permission_class="api")
    for group in Groups.objects.all():
        random_permission = random.sample(list(permissions), int(len(permissions)/2))
        for p in random_permission:
            p = model_to_dict(p)
            group.permissions[p["key"] + ":" + p["method"]] = 1
            group.save()

if __name__ == '__main__':
    # import_permissions(1)
    # import_groups(1)
    # set_groups_permission(1)
    # print(set_company_groups(1))
    pass


# -*- coding: utf-8 -*-
import random, time
from lcyframe.libs.beanstalk_route import BsRoute
from base import BaseModel

"""
目前仅支持监听default tube；所有任务都在此通道
self.beanstalk.DemoEvents.event({"DemoEvents": 1})
"""

@BsRoute.tasks
class DemoEvents(BaseModel):
    tube = "default"

    @classmethod
    def event(cls, job):
        """
        返回False：job不会删除
        返回非False：job会被删除
        :param x:
        :return:
        """
        print(job)
        return True



# coding=utf-8
from .libs.mqtt_route import MqttTask
from .libs.mqtt_route import ReadMqtt

class MqttWorker(ReadMqtt):
    def __init__(self, **kwargs):
        super(MqttWorker, self).__init__(**kwargs["mqtt_config"])
        self.ROOT = kwargs["ROOT"]
        self.mqtt_config = kwargs["mqtt_config"]
        self.workers_path = self.mqtt_config.get("workers_dir", ["works/mqtt"])
        self.workers = MqttTask.get_workers(self.ROOT, self.workers_path)

if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.DEBUG)
    kwargs = {"mqtt_config":
        {
          "host": "localhost",
          "port": 1883,
          "client_id_prefix": "yyq",
          "username": "admin",
          "password": "123456",
          "msg_pwd": 123456

        }
    }
    wk = MqttWorker(**kwargs)
    wk.start()



import datetime, time, random
import logging
from traceback import format_exc
import threading

try:
    import pymysql
    from dbutils.pooled_db import PooledDB, TooManyConnections
    from dbutils.persistent_db import PersistentDB
except:
    print("not found model pymysql")


class PyMysqlPooledDB(object):
    """
    链接池
    用于一般应用服务
    """
    pool = None

    def __init__(self, **kwargs):
        """
        :param host:数据库ip地址
        :param port:数据库端口
        :param database:库名
        :param user:用户名
        :param password:密码
        :param charset:字符编码
        :param read_timeout: The timeout for reading from the connection in seconds (default: None - no timeout)
        :param write_timeout: The timeout for writing to the connection in seconds (default: None - no timeout)
        :param maxconnections:创建连接池的最大数量
        :param mincached:连接池中空闲连接的初始数量，0表示启动是不建立连接
        :param maxcached:连接池中空闲连接的最大数量，0不受限制
        :param maxusage:单个连接的最大重复使用次数，0 无限重复使用
        :param maxshared:共享连接的最大数量, 0禁止共享链接,由于pymysql.threadsafety=1，不是线程安全的，所以禁止线程间共享连接
        :param blocking:超过最大连接数量时，True阻塞等待连接数量下降，为false直接报错处理
        :param setsession:optional list of SQL commands that may serve to prepare
            the session, e.g. ["set datestyle to ...", "set time zone ..."]
        :param reset:how connections should be reset when returned to the pool
            (False or None to rollback transcations started with begin(),
            True to always issue a rollback for safety's sake)
        :param Autocommit mode. None means use server default. (default: False)
        自动提交事务，每执行一条sql语句，事务提交一次；关闭后，需手动调用commit提交事务；建议设置为False。mysql服务端上的模式不需要改动(其默认为True)
        更多参数，参考：./site-packages/pymysql/connections.py (class Connection)
        >>> connetion()
            begin()     开启事务
            commit()    提交事务
            cursor(cursor=None)创建一个游标以执行语句
            ping(reconnect=True)检查连接是否存活，并重新发起连接
            rollback()  回滚事务
            close()     关闭连接
            select_db(db)选择数据库
            show_warnings()查看warning信息
        >>> cursor()
            fetchone()      获取下一行数据。
            fetchall()      获取所有数据。
            fetchmany(size=None)    获取几行数据。
            read_next()     获取下一行数据。
            callproc()      用来调用存储过程。
            mogrify()       参数化查询，防止SQL注入。
            scroll(num,mode)移动游标位置。
            close()         关闭
        """
        kwargs.setdefault("read_timeout", 300)
        kwargs.setdefault("write_timeout", 300)
        if not self.pool:
            cursorclass = {1: pymysql.cursors.DictCursor,
                           2: pymysql.cursors.Cursor}[kwargs.pop("cursorclass", 1)]
            self.pool = PooledDB(pymysql,  # 默认pymysql.threadsafety=1，是非线性安全的，线程可以共享模块，但不能共享连接。
                                 host=kwargs.pop("host", "127.0.0.1"),
                                 port=kwargs.pop("port", 3306),
                                 user=kwargs.pop("user", 'root'),
                                 password=str(kwargs.pop("password", "")),
                                 database=kwargs.pop("database", "test"),
                                 mincached=kwargs.pop("mincached", 1),
                                 maxcached=kwargs.pop("maxcached", 0),
                                 maxconnections=kwargs.pop("maxconnections", 32),
                                 maxusage=kwargs.pop("maxusage", 100),
                                 blocking=kwargs.pop("blocking", True),
                                 setsession=kwargs.pop("setsession", None),
                                 reset=kwargs.pop("reset", True),
                                 charset=kwargs.pop("charset", "utf8mb4"),
                                 cursorclass=cursorclass,
                                 **kwargs
                                 )
        self.shareable = kwargs.pop("shareable", True)  # pymsql 非线程安全，设置True也不起作用
        self.lock = threading.Lock()

    def get_connection(self):
        self.lock.acquire()
        # 从连接池中，取出一个连接,如果设置了ping，会自动重连:conn._ping_check()
        conn = self.pool.connection(self.shareable)  # True, 允许线程共享链接，前提是基础API包(这里指pymysql包,非线程安全)是线程安全的。
        # print(self.pool._connections)  # 当前使用中的连接数（未放回池内）
        self.lock.release()
        return conn

    def select_one(self, sql, params=None, safe_conn=None):
        """
        :param sql: qsl语句
        :param param: sql参数
        :param safe_conn 启动事务时，必传。唯一链接，防止多线程下交叉使用，提交数据
        :example:
            select_one("select * from demo where name=%s", params=(2, ))
        """
        conn = safe_conn or self.get_connection()
        rows, cursor = self.__execute(conn, sql, params)
        result = cursor.fetchone()
        """:type result:dict"""
        result = self.__dict_datetime_obj_to_str(result)
        return result

    def select_many(self, sql, params=None, safe_conn=None):
        """
        :param sql: qsl语句
        :param param: sql参数
        :return: 结果数量和查询结果集
        :example:
            self.select_many('select * from demo where name=%s', (2,))
        """
        conn = safe_conn or self.get_connection()
        count, cursor = self.__execute(conn, sql, params)
        result = cursor.fetchall()
        """:type result:list"""
        [self.__dict_datetime_obj_to_str(row_dict) for row_dict in result]
        return result

    def insert(self, table, insert_data, many=False, new=True, safe_conn=None):
        """
        from pymysql.converters import escape_string
        content: escape_string(content)
        :param table:
        :param insert_data  type:[{"field1": 1},{"field2": 2}]
        :param many  type: true 批量写入（last_insert_id()得到的值不准确）
                           false 逐条写入（last_insert_id()得到的值为准确值,需要获得最新的id值，请使用逐条写入）
                           last_insert_id(): 当前连接，整个数据库内，上一次insert事务提交后，生成的最后一个自增key的值,
                           若连接断开，则值为0
        :param new  type: 若id为自增键时，True 返回最新记录，False返回影响条数
        :return:count 1 影响的行数
        """
        conn = safe_conn or self.get_connection()
        count = 0
        last_insert_ids = []
        insert_datas = insert_data if isinstance(insert_data, (list, tuple)) else [insert_data]
        try:
            columns = ','.join(list(insert_datas[0].keys()))
            params = [tuple(data.values()) for data in insert_datas]
            values = (len(params[0]) * "%s,").rstrip(",")
            sql = "insert into " + table + " (" + columns + ") values (" + values + ")"
            if many:
                count, cursor = self.__execute(conn, sql, params, True)
                result = count
            else:
                for item in params:
                    rows, cursor = self.__execute(conn, sql, item)
                    count += rows
                    if new:
                        # 返回当前连接上一次执行insert语句后，得到的全局自增id值:select last_insert_id()
                        rows, cursor = self.__execute(conn, "select last_insert_id() as id")
                        result = cursor.fetchone()
                        if result:
                            last_insert_ids.append(result["id"])
        except Exception as e:
            self.rollback(conn)
            raise e
        else:
            if new:
                if not isinstance(insert_data, (list, tuple)):
                    sql = f'select * from {table} where id = %s'
                    result = self.select_one(sql, last_insert_ids[0], safe_conn=conn)
                else:
                    sql = f'select * from {table} where id in %s'
                    result = self.select_many(sql, [tuple(last_insert_ids)], safe_conn=conn)
            self.__commit(conn)
            self.close(conn, cursor)
            return result

    def inserts(self, table, insert_data, new=False, safe_conn=None):
        """
        批量写入
        new： True 返回新写入的数据列表（写入为逐条写入，非原子性） False不返回最新数据，只返回影调条数（原子写入）
        """
        if new:
            results = self.insert(table, insert_data, many=False, new=True, safe_conn=safe_conn)
        else:
            results = self.insert(table, insert_data, many=True, new=False, safe_conn=safe_conn)
        return results

    def update(self, sql, params=None, safe_conn=None):
        """
        :param sql:
        :param params:
        :return:
        :example:
            # update all data
            self.update('update demo set name=22 where name=%s', (2,))
            # update first one with order asc
            self.update('update demo set name=2 where name=%s order by id limit 1', (22,))
        """
        conn = safe_conn or self.get_connection()
        count, cursor = self.__execute(conn, sql, params)
        self.__commit(conn)
        self.close(conn, cursor)
        return count

    def delete(self, sql, params=None, safe_conn=None):
        """
        :param sql:
        :param params:
        :return:
        :example:
            # delete all data
            self.delete('delete from demo where name=%s', (22,))
            # delete first one with order asc
            self.delete('delete from demo where name=%s order by id limit 1', (22,))
        """
        conn = safe_conn or self.get_connection()
        count, cursor = self.__execute(conn, sql, params)
        self.__commit(conn)
        self.close(conn, cursor)
        return count

    def next_id(self, safe_conn=None):
        """
        下一个自增id值
        """
        conn = safe_conn or self.get_connection()
        count, cursor = self.__execute(conn, "select last_insert_id() as id")
        result = cursor.fetchone()
        return result["id"]

    def query_sql(self, sql, params=None, safe_conn=None):
        """
        查询数据
        sql不要拼接或采用format()，请用传参方式，防止注入

        If args is a list or tuple, %s can be used as a placeholder in the query.
        If args is a dict, %(name)s can be used as a placeholder in the query.

        :param sql:
        :param param:
        :example::
            data = query_sql("select * from demo where name=%s and id=%s", (1, 2))
            data = query_sql("select * from demo where name=%(name)s", {"name": 2})
            data = query_sql("select * from demo where name=name", {"name": 2})
        :return:

        """
        conn = safe_conn or self.get_connection()
        rows, cursor = self.__execute(conn, sql, params)
        result = cursor.fetchall()
        self.close(conn, cursor)
        return result or []

    def execute(self, sql, params=None, safe_conn=None):
        """
        提交数据
        原生执行，sql不要拼接.format，用传参方式，防止注入
        常用于：insert、delete、update
        :param sql:
        :param param: [1, 2]、[(1, 2), (3, 4)]
        :return:
        """
        conn = safe_conn or self.get_connection()
        count, cursor = self.__execute(conn, sql, params)
        self.__commit(conn)
        self.close(conn, cursor)
        return count

    def begin(self):
        """
        开启事务:
        1、关闭mysql自动提交配置；set autocommit=0
        2、执行事务逻辑
        3、需显示调用self.commit提交事务
        开启事务过程中若连接断开，禁止重新获取conn，只能报错后rollback处理，防止conn对象不一致
        该方法将返回一个链接（事务安全链接,不和其他线程共享），
        需以参数形式传给后续的所有操作，否则在多线程情况下出现线程安全问题，
        业务需要捕获处理异常，结束事务，已回收释放链接。
        ref: 您需要通过调用begin（）方法显式启动事务。这可以确保连接不会与其他线程共享，透明的重新打开将被挂起，直到事务结束，并且连接将在返回到连接池之前回滚。
        """
        conn = self.get_connection()
        conn.begin()
        # 设置事务标识
        setattr(conn, "is_transcations_running", True)
        return conn

    
    def commit(self, conn):
        """
        结束事务：必须先调用self.begin开启事务
        提交事务，回收连接
        """
        if not hasattr(conn, "is_transcations_running"):
            raise Exception("请先显式调用begin()开启事务")
        del conn.is_transcations_running
        conn.commit()
        conn.close()

    def rollback(self, conn):
        """
        回滚
        :return:
        """
        conn.rollback()
        conn.close()

    def close(self, conn, cursor=None):
        """
        将连接放回连接池
        """
        try:
            if hasattr(conn, "is_transcations_running") and conn.is_transcations_running:
                pass
            else:
                cursor.close()
                conn.close()
        except Exception as e:
            raise e

    def parse_set_condition_value(self, **params):
        sql_columns = ""
        for k in params.keys():
            sql_columns += k + "=%s,"
        sql_columns = sql_columns.rstrip(",")
        sql_values = params.values()
        return sql_columns, list(sql_values)
    
    def __execute(self, conn, sql, params=None, many=False):
        """
        业务需要过滤敏感字符，防止sql注入
        1、禁止sql采用参数拼接
        2、过滤group by内非法字符
        3、过滤order by内非法字符
        """
        conn._ping_check()  # 执行前，重连当前连接,确保连接有效。若处于事务中，则无法重连，连接直接关闭
        cursor = conn.cursor()
        try:
            if many:
                rows = cursor.executemany(sql, params)
            else:
                rows = cursor.execute(sql, params)
            return rows, cursor
        except pymysql.err.InterfaceError as e:
            logging.error("DB连接实例无法连通，请检查mysql是否启动")
            logging.debug("正常尝试重连mysql...")
            conn._ping_check()
            self.rollback(conn)
            raise e
        except TooManyConnections as e:
            logging.error("Pool连接池内无空闲连接")
            logging.error(format_exc())
            self.rollback(conn)
            raise e
        except Exception as e:
            logging.error(format_exc())
            self.rollback(conn)
            raise e

    @staticmethod
    def __dict_datetime_obj_to_str(result_dict):
        """把字典里面的datatime对象转成字符串"""
        if result_dict:
            result_replace = {k: v.__str__() for k, v in result_dict.items() if isinstance(v, datetime.datetime)}
            result_dict.update(result_replace)
        return result_dict

    def __commit(self, conn):
        """
        提交数据，
        """
        #  若conn.is_transcations_running=True时，不执行提交。由开启事务的业务流程主动调用end()提交事务
        if hasattr(conn, "is_transcations_running") and conn.is_transcations_running:
            pass
        else:
            conn.commit()

class PyMysqlPersistentDB(PyMysqlPooledDB):
    """
    用于多线程编程
    线程专用链接模式
    """
    pool = None

    def __init__(self, **kwargs):
        cursorclass = {1: pymysql.cursors.DictCursor,
                       2: pymysql.cursors.Cursor}[kwargs.pop("cursorclass", 1)]
        kwargs.setdefault("read_timeout", 300)
        kwargs.setdefault("write_timeout", 300)
        if not self.pool:
            self.pool = PersistentDB(
                creator=pymysql,  # 使用链接数据库的模块# 默认pymysql.threadsafety=1是非线程安全。线程可以共享模块，但不能共享连接。
                host=kwargs.pop("host", "127.0.0.1"),
                port=kwargs.pop("port", 3306),
                user=kwargs.pop("user", 'root'),
                password=str(kwargs.pop("password", "")),
                database=kwargs.pop("database", "test"),
                maxusage=kwargs.pop("maxusage", None),  # 一个链接最多被重复使用的次数，None表示无限制
                setsession=kwargs.pop("setsession", []),  # 开始会话前执行的命令列表。如：["set datestyle to ...", "set time zone ..."]
                closeable=kwargs.pop("closeable", False),
                # 如果为False时， conn.close() 实际上被忽略，供下次使用，主线程关闭时才会关闭链接。如果为True时， conn.close()则关闭链接，那么再次调用pool.connection时就会报错。（pool.steady_connection()可以获取一个新的链接）
                threadlocal=kwargs.pop("threadlocal", None),  # 本线程独享值得对象，用于保存链接对象，如果链接对象被重置
                charset=kwargs.pop("charset", "utf8mb4"),
                cursorclass=cursorclass,
                **kwargs
            )
        self.shareable = kwargs.pop("shareable", False)
        self.lock = threading.Lock()




#!/usr/bin/env python
import sys
import asyncio
import websockets
import inspect

from base import BaseModel
from lcyframe import yaml2py
from lcyframe import ServicesFindet
from lcyframe.websocket_server import WebSocketWorker
from context import InitContext

# args = [item.split("=")[-1] for item in sys.argv if item.startswith("--config")]
sys.argv = sys.argv[:1]
config = InitContext.get_context()

conns = set()

# async def handler(websocket):
#     conns.add(websocket)
#     ws = None
#     try:
#         async for message in websocket:
#             # await websocket.send(message)
#
#             # 全体发送
#             for ws in conns:
#                 if ws:
#                     try:
#                         await ws.send(message)
#                     except websockets.ConnectionClosedOK as e:
#                         print("连接已断开：%s:%s" % ws.remote_address[:2])
#                         if ws in conns:
#                             conns.remove(ws)
#     except Exception as e:
#         if ws in conns:
#             conns.remove(ws)
#         print(str(e))   # websockets.ConnectionClosedOK


async def handler(websocket):
    conns.add(websocket)

    while True:
        try:
            message = await websocket.recv()

            # 循环发送速度过快，容易造成网络阻塞，导致资源暂满网络缓冲区，可以用背压，没100毫秒发送一批数据
            # for ws in conns:
            #     await ws.send(message)

            # 更优的方式实现批量发送
            if message:
                websockets.broadcast(conns, message)
        except Exception as e:
            if websocket in conns:
                conns.remove(websocket)
                print("连接已断开：%s:%s" % websocket.remote_address[:2])
            else:
                # websocket.close_connection()    # 关闭当前线程连接
                websocket.close()                 # 当连接断开，或者ping不他不通时(ping_interval)，主动关闭当前线程连接
            break

async def main():
    # 每一个连接都单独一个线程
    async with websockets.serve(handler, "localhost", 8765,
                                ping_interval=20,   # 一旦连接打开，每秒钟发送一个Ping帧ping_interval 。这用作保活。它有助于保持连接打开，尤其是在非活动连接上存在短暂超时的代理的情况下。设置ping_interval为None禁用此行为。
                                # create_protocol=websockets.basic_auth_protocol_factory(
                                #     realm="my dev server",
                                #     credentials=("hello", "iloveyou"),
                                # )
                                ):
        await asyncio.Future()  # run forever

if __name__ == "__main__":
    # asyncio.run(main())

    websocket = WebSocketWorker(**config)
    yaml2py.impmodule(BaseModel, "model")  # 换为载入model父类，切worker的函数继承了model父类，所以可以使用所有的model子类
    ServicesFindet(websocket, BaseModel)(config)  # 获取全站所有连接库，赋值给BaseModel，继承了BaseModel的worker内，含有mqtt生产者连接，可以使用生产者创建新的任务
    websocket.model = BaseModel.model
    websocket.start()


# -*- coding:utf-8 -*-
from falcon.status_codes import HTTP_BAD_REQUEST, HTTP_NOT_FOUND, HTTP_FORBIDDEN


class ApiError(Exception):
    code = 1
    code_name = 'Api Runtime Error'
    message = 'Runtime api error occurred.'
    status = HTTP_BAD_REQUEST

    def __init__(self, message=None, code=None, code_name=None, *args, **kwargs):
        super(ApiError, self).__init__(*args, **kwargs)
        if message:
            self.message = message
        if code is not None:
            self.code = code
        if code_name is not None:
            self.code_name = code_name

    def to_dict(self, *args, **kwargs):
        print({
            'err_code': self.code,
            'err_name': self.code_name,
            'message': self.message
        })
        return {
            'err_code': self.code,
            'err_name': self.code_name,
            'message': self.message
        }

# 非法请求
class ErrorInvalid(ApiError):
    """
    非法请求
    """
    code = 2
    code_name = 'invalid.'
    message = 'invalid request.'
    zh_message = '非法请求'

# 非法跨域
class ErrorCorsUri(ApiError):
    """
    非法请求
    """
    code = 3
    code_name = 'ErrorCorsUri'
    message = 'ErrorCorsUri'
    zh_message = '来自不授信的跨域请求'

# 参数类错误40x
class ErrorMissingArgument(ApiError):
    """
    参数缺失或错误
    """
    code = 401
    code_name = 'Missing Argument'
    zh_message = '参数缺失'


class ErrorArgumentType(ApiError):
    """
    参数类型错误
    """
    code = 402
    code_name = 'Argument Type Error'
    message = 'Argument Type Error'
    zh_message = '参数类型错误'

class ErrorArgumentValue(ApiError):
    """
    参数值错误
    """
    code = 403
    code_name = 'Argument Value Error'
    zh_message = '参数值不在允许范围'


# 返回类错误50x
class ErrorResponse(ApiError):
    """
    返回结构
    """
    code = 501
    code_name = 'Response Key Error!'
    zh_message = '返回结构与定义的不一致'


# 授权类错误 60x
class ErrorTokenInvalid(ApiError):
    """
    无效token
    """
    code = 601
    code_name = 'Invalid_Access_Token'
    message = 'Access token invalid.'
    zh_message = '无效的token'
    status = HTTP_FORBIDDEN

class ErrorTokenExpireInvalid(ApiError):
    """
    token已过期
    """
    code = 602
    code_name = 'Invalid_Access_Token'
    message = 'Access token expired.'
    zh_message = 'token已过期'
    status = HTTP_FORBIDDEN

class ErrorRefreshTokenInvalid(ApiError):
    code = 603
    code_name = 'Invalid_Refresh_Token'
    message = 'Refresh token invalid, refresh access-token failed.'
    zh_message = ''
    status = HTTP_FORBIDDEN

class ErrorSignInvalid(ApiError):
    code = 604
    code_name = 'ErrorSignInvalid'
    message = '非法的sign'
    status = HTTP_FORBIDDEN

class ErrorExpireQuery(ApiError):
    code = 605
    code_name = 'ErrorExpireQuery'
    message = '该请求被系统判断为无效请求'
    status = HTTP_FORBIDDEN

class ErrorDuplicateQuery(ApiError):
    code = 606
    code_name = 'ErrorDuplicateQuery'
    message = '该请求被系统判断为重复请求'
    status = HTTP_FORBIDDEN

class ErrorCommon(ApiError):
    """
    通用类，无特定使用界限
    """
    code = 700
    code_name = 'ErrorCommon'
    message = 'ErrorCommon'
    zh_message = ''

class ErrorNotData(ApiError):
    """
    通用类：数据不存在
    """
    code = 701
    code_name = 'ErrorNotData'
    message = 'ErrorNotData'
    zh_message = '数据不存在'

class ErrorStateError(ApiError):
    """
    通用类：状态不符合
    """
    code = 702
    code_name = 'ErrorStateError'
    message = 'ErrorStateError'
    zh_message = '状态错误'

# 权限类错误90x
class ErrorNoApiPermission(ApiError):
    """
    该账号没有调用该接口的权限
    """
    code = 901
    code_name = 'No_Api_Permission'
    message = 'No api permission'
    zh_message = '没有调用该接口的权限'


class ErrorNoRolePermission(ApiError):
    """
    没有操作参数中角色或职位的权限
    """
    code = 902
    code_name = 'No_Role_Permission'
    message = 'No role permission'
    zh_message = '没有操作参数中角色或职位的权限'


class UploadError(ApiError):
    """
    格式错误
    """
    code = 903
    code_name = 'FileFormatError'
    message = '不支持的格式或文件过大'




import base64
import random
import string
from io import BytesIO
from pathlib import Path
from random import randint, choice
from PIL import ImageFont, ImageFilter, Image, ImageColor, ImageDraw
from typing import List, Tuple, Any, Union, Optional


class VFCode:

    def __init__(
            self,
            width: int = 200,
            height: int = 80,
            fontsize: int = 50,
            font_color_values: Optional[List[str]] = None,
            font_background_value: Optional[str] = None,
            draw_dots: bool = False,
            dots_width: int = 1,
            dots_count: int = 50,
            draw_lines: bool = True,
            lines_width: int = 3,
            lines_count: int = 6,
            lines_background_value: Optional[str] = None,
            mask: bool = False,
            font: Optional[str] = None
    ):
        """
        初始化参数
        :param width: 图片宽度
        :param height: 图片高度
        :param fontsize: 字体尺寸
        :param font_color_values: 字体颜色值
        :param font_background_value: 背景颜色值
        :param draw_dots: 是否画干扰点
        :param dots_width: 干扰点宽度
        :param draw_lines: 是否画干扰线
        :param lines_width: 干扰线宽度
        :param mask: 是否使用磨砂效果
        :param font: 字体 内置可选字体 arial.ttf calibri.ttf simsun.ttc
        """
        self.code = None
        self._img = None
        self.width = width
        self.height = height
        self.fontsize = fontsize
        # linux add fonts file /usr/share/fonts/arial.ttf
        if font is None:
            self.font = ImageFont.truetype(str(Path(__file__).resolve().parent) + '/fonts/arial.ttf', fontsize)
        else:
            try:
                self.font = ImageFont.truetype(str(Path(__file__).resolve().parent) + '/fonts/%s' % font, fontsize)
            except:
                self.font = ImageFont.truetype(font, fontsize)
        self.draw_dots = draw_dots
        self.dots_width = dots_width
        self.dots_count = dots_count
        self.draw_lines = draw_lines
        self.lines_width = lines_width
        self.lines_count = lines_count
        self.lines_background_value = lines_background_value
        self.mask = mask

        self.symbol = None
        self.result = None

        if font_color_values:
            self.font_colors = [ImageColor.getcolor(value, 'RGBA') for value in font_color_values]
        else:
            color_values = [
                # '#ffffff',
                '#000000',
                '#3e3e3e',
                '#ff1107',
                '#1bff46',
                '#ffbf13',
                '#235aff'
            ]
            self.font_colors = [ImageColor.getcolor(value, 'RGBA') for value in color_values]

        if font_background_value:
            self.font_background = ImageColor.getcolor(font_background_value, 'RGBA')
        else:
            self.font_background = ImageColor.getcolor('#ffffff', 'RGBA')

        # 移除字体颜色和背景颜色相同的值
        if self.font_background in self.font_colors:
            self.font_colors.remove(self.font_background)

    def random_color(self) -> Tuple:
        return choice(self.font_colors)

    def img_blur(self, img: Image) -> None:
        width, height = img.size
        draw = ImageDraw.Draw(img)

        if self.draw_lines:
            for i in range(self.lines_count):
                x1 = randint(0, width)
                x2 = randint(0, width)
                y1 = randint(0, height)
                y2 = randint(0, height)
                draw.line((x1, y1, x2, y2), fill=self.lines_background_value or self.random_color(), width=self.lines_width)

        if self.draw_dots:
            for i in range(self.dots_count):
                draw.point([randint(0, width), randint(0, height)], fill=self.random_color())
                x = randint(0, width)
                y = randint(0, height)
                draw.arc((x, y, x + 4, y + 4), 0, 90, fill=self.random_color(), width=self.dots_width)

    def _gen_txt_img(self, text: str) -> Image.Image:
        img = Image.new('RGBA', (self.fontsize + 5, self.fontsize + 5))
        text_draw = ImageDraw.Draw(img)
        text_draw.text((0, 0), text, fill=self.random_color(), font=self.font)

        if self.symbol and self.result:
            if text == self.symbol:
                return img
            img = img.rotate(angle=randint(-30, 30), expand=False)
        else:
            img = img.rotate(angle=randint(-30, 90), expand=False)
        return img

    def generate(self, code) -> None:
        self.code = code
        self._img = Image.new(
            'RGBA',
            (self.width, self.height),
            color=self.font_background
        )
        # blur image
        self.img_blur(self._img)
        if self.symbol and self.result:
            deal_code = code.split(self.symbol)
            deal_code.insert(1, self.symbol)
        else:
            deal_code = code
        base = int(self.width / len(deal_code))
        for pos, c in enumerate(deal_code):
            txt_img = self._gen_txt_img(c)
            roffset = randint(-10, 10)

            # set the position and paste the code image
            x, y = pos * base + roffset, int((self.height - txt_img.height) / 2)

            if x < 0:
                x = 0
            if x > self.width:
                x = self.width
            self._img.paste(txt_img, box=(x, y), mask=txt_img)

        if self.mask:
            # ImageFilter.GaussianBlur
            self._img = self._img.filter(ImageFilter.GaussianBlur(radius=1))

    def generate_digit(self, length: int = 5) -> None:
        self.generate(''.join(str(random.randrange(10 ** (length - 1), 10 ** length))))

    def generate_alpha(self, length: int = 5) -> None:
        code_list = []
        for i in range(length):
            code_list.append(choice(string.ascii_letters))

        self.generate(''.join(code_list))

    def generate_mix(self, length: int = 5) -> None:
        code_list = []
        for i in range(length):
            code_list.append(choice(string.digits + string.ascii_letters))

        self.generate(''.join(code_list))

    def generate_op(self, symbol: str = 'x') -> None:
        self.symbol = symbol
        left_digit = random.randrange(10, 100)
        right_digit = random.randrange(10, 100)
        while left_digit <= right_digit:
            right_digit //= 2
        if symbol == '+':
            self.result = left_digit + right_digit
        elif symbol == '-':
            self.result = left_digit - right_digit
        elif symbol == 'x':
            left_digit = random.randrange(1, 10)
            right_digit = random.randrange(1, 10)
            self.result = left_digit * right_digit
        else:
            raise ValueError('only suppo +-x')
        self.generate('%s%s%s' % (left_digit, symbol, right_digit))

    def get_img_bytes(self, fm='png') -> bytes:
        binary_stream = BytesIO()
        self._img.save(binary_stream, fm)
        return binary_stream.getvalue()

    def get_img_base64(self, fm='png') -> Tuple[Any, Union[str, Any]]:
        b64_img_prefix = 'data:image/png;base64,'
        base64_img = b64_img_prefix + base64.b64encode(self.get_img_bytes(fm)).decode()
        if self.symbol and self.result:
            return str(self.result), base64_img
        else:
            return self.code, base64_img

    def save(self, filename: str = None, fm='png') -> None:
        if filename is None:
            self._img.save(self.code + '.' + fm, fm)
        else:
            self._img.save(filename, fm)


if __name__ == '__main__':
    vc = VFCode(
        width=200,                       # 图片宽度
        height=80,                       # 图片高度
        fontsize=50,                     # 字体尺寸
        font_color_values=[
            '#ffffff',
            '#000000',
            '#3e3e3e',
            '#ff1107',
            '#1bff46',
            '#ffbf13',
            '#235aff'
        ],                                # 字体颜色值
        font_background_value='#ffffff',  # 背景颜色值
        draw_dots=False,                  # 是否画干扰点
        dots_width=1,                     # 干扰点宽度
        draw_lines=True,                  # 是否画干扰线
        lines_width=3,                    # 干扰线宽度
        mask=False,                       # 是否使用磨砂效果
        font='arial.ttf'                  # 字体 内置可选字体 arial.ttf calibri.ttf simsun.ttc
    )
    # 验证码类型
    # 自定义验证码
    # vc.generate('abcd')

    # 数字验证码（默认5位）
    # vc.generate_digit()
    # vc.generate_digit(4)

    # 字母验证码（默认5位）
    # vc.generate_alpha()
    # vc.generate_alpha(5)

    # 数字字母混合验证码（默认5位）
    # vc.generate_mix()
    # vc.generate_mix(6)

    # 数字加减验证码（默认加法）
    vc.generate_op()
    # 数字加减验证码（加法）
    # vc.generate_op('+')
    # 数字加减验证码（减法）
    # vc.generate_op('-')

    # 图片字节码
    # print(vc.get_img_bytes())
    # 图片base64编码
    print(vc.get_img_base64())
    # 保存图片
    vc.save()

    # 验证码
    print(vc.code)


# coding=utf-8
"""权限"""
import logging
import time, random
from bson.objectid import ObjectId
from lcyframe.libs.singleton import MongoCon
from lcyframe.libs import errors


class BitPermission(object):
    PERMISSION_TABLE = "permission_group"
    PERMISSION_CLASS_TABLE = "permission_class"

    CURRENT_PERMISSION_GROUP = None
    CURRENT_PERMISSION_CLASS = None
    __VERSION__ = 0
    MAX_NUMS = 63
    IS_DEBUG = False
    default_group_config = None
    default_class_config = None
    mongo = MongoCon._database

    @classmethod
    def initial(cls):
        if cls.mongo is None:
            cls.mongo = MongoCon._database
            if cls.mongo is None:
                raise Exception("you must get a mongo isinstance!")

        if cls.CURRENT_PERMISSION_GROUP is None:
            cls.__load_permission_group()
        if cls.CURRENT_PERMISSION_CLASS is None:
            cls.__load_permission_class()

        return cls

    @classmethod
    def bit_mp_obj(cls, name, sort, available=True):
        return {"name": name,
                "sort": sort,
                "desc": "",
                "available": available}

    @classmethod
    def gen_bit_mp(cls, nums):
        """
        生成位权限对象
        :param nums:  权限位数
        :return:
        """
        nums = max(nums, cls.MAX_NUMS)

        bit_mp = {}
        for i in range(nums):
            bit_mp[str(i)] = cls.bit_mp_obj("权限%s" % i, i)
        return bit_mp

    @classmethod
    def gen_class_obj(cls, k, nums=MAX_NUMS):
        """
        生成权限类别
        :return:
        """
        return {
            "_id": str(ObjectId()),
            "key": int(k),
            "available": True,
            "name": "权限类别:%s" % str(k),
            "bit_mp": cls.gen_bit_mp(nums),
            "create_at": int(time.time()),
            "desc": ""
        }

    @classmethod
    def gen_permission_class(cls, *class_obj, **kwargs):
        """
        生成多个权限类别和权限
        "config": [
                {
                "_id": "f3245d",
                "key": 1,   # 主键
                "available": True,
                "name": "用户管理权限类",
                "bit_mp": {
                    "0": {"name": "可创建用户", "available": True},   # 0位
                    "1": {"name": "可创建用户", "available": True},   # 1位
                    ...,
                    "62": {"name": "可创建用户", "available": True}   # 63位
                },
                ...
        ]
        :param ids:  权限类别ids
        :param nums:  小于64  每个类别最多可设置63种权限
        :return:
        """
        class_obj = class_obj or cls.default_class_config

        d = []

        if class_obj:
            if cls.mongo[cls.PERMISSION_CLASS_TABLE].find_one():
                raise Exception("the permission class is exists, you can create new obj.")

            if not cls.check_permission_class(class_obj):
                raise Exception("the permission config is error.")

            for _class in class_obj:
                _class.setdefault("_id", str(ObjectId()))
                d.append(_class)
        else:
            for k, key in enumerate(kwargs["keys"]):
                d.append(cls.gen_class_obj(k + 1, kwargs.get("nums", cls.MAX_NUMS)))

        cls.mongo[cls.PERMISSION_CLASS_TABLE].insert(d)
        return d

    @classmethod
    def gen_perminssion_group(cls, config=None):
        """
        角色组配置表
        :param config:  [
                            {
                                "gid": 1,   # 权限组ID，大小与父子关系没有必然联系
                                "pid": 1,   # 父子关系，1 > 2 > 3，值越小层级越高
                                "available": True,
                                "name": "角色组1",
                                "permission_class": {
                                        "0": 13243,   # 权限类key=xxx拥有的权限值为13243,
                                        "1": 54432,   # 权限类key=yyy拥有的权限值为13243
                                }
                            },
                            ...
                            {
                                "gid": 2,   # 权限组ID，大小与父子关系没有必然联系
                                "pid": 2,   # 父子关系，1 > 2 > 3，值越小层级越高
                                "available": True,
                                "name": "角色组5",
                                "permission_class": {
                                        "0": 13243,   # 权限类key=xxx拥有的权限值为13243,
                                        "1": 54432,   # 权限类key=xxx拥有的权限值为13243
                                }
                            }
                        ]
        :return:
        """
        config = config or cls.default_group_config
        if config and not cls.check_permission_group(config):
            return False

        if not config:
            permission_group_list = []
            classes = cls.gen_permission_class(keys=[ObjectId(), ObjectId()])
            permission_class = {}
            for g in classes:
                permission_class[str(g["key"])] = 0

            for i in range(1, 6):
                permission_group_list.append({
                    "_id": str(ObjectId()),
                    "gid": i,
                    "pid": i,
                    "available": True,
                    "name": "角色%d" % i,
                    "permission_class": permission_class,
                    "create_at": int(time.time()),
                    "desc": ""
                }
                )

            config = permission_group_list

        cls.mongo[cls.PERMISSION_TABLE].insert(config)
        return config

    @classmethod
    def check_permission_group(cls, group):
        """检测角色组配置文件
        {
            "gid": 1,
            "available": True,
            "name": "角色组1",
            "permission_class": {
                    "1": 13243,   # 权限类key=xxx拥有的权限值为13243,
                    "2": 54432,   # 权限类key=yyy拥有的权限值为13243
            }
        }
        """

        if not isinstance(group, list):
            group = [group]

        for g in group:
            try:
                assert type(g["gid"]) == int
                assert type(g["available"]) == bool
                assert type(g["name"]) in [str, unicode]
                assert type(g["desc"]) in [str, unicode]
                assert type(g["permission_class"]) == dict
                for k, v in g["permission_class"].items():
                    assert type(k) == str
                    assert type(v) == int
                return True
            except:
                return False

    @classmethod
    def check_permission_class(cls, _class):
        """检测权限类配置文件
        {
            "_id": "f3245d",
            "key": "1",   # 主键
            "available": True,
            "name": "用户管理权限类",
            "bit_mp": {
                "0": {"name": "可创建用户", "available": True},   # 0位
                "1": {"name": "可创建用户", "available": True},   # 1位
                ...,
                "62": {"name": "可创建用户", "available": True}   # 63位
            }
        """

        if type(_class) not in (tuple, list):
            _class = [_class]

        for g in _class:
            try:
                if "_id" in g:
                    assert type(g["_id"]) == str

                assert type(g["key"]) == int
                assert type(g["name"]) in [str, unicode]
                assert type(g["desc"]) in [str, unicode]
                assert type(g["bit_mp"]) == dict
                for k, v in g["bit_mp"].items():
                    assert int(k) >= 0 and int(k) <= cls.MAX_NUMS
                    assert type(v) == dict
                return True
            except:
                return False

    @classmethod
    def __load_permission_group(cls):
        """载入角色组"""
        mp = {"group": {}, "parent": {}}

        group = list(cls.mongo[cls.PERMISSION_TABLE].find())

        if not group:
            group = cls.gen_perminssion_group()

        for g in group:
            g.pop("_id", None)
            mp["group"][int(g["gid"])] = g  # 角色组mp
            mp["parent"][int(g["pid"])] = g  # 父子组mp， 获取某个gid的父级角色时使用

        cls.CURRENT_PERMISSION_GROUP = mp
        return cls.CURRENT_PERMISSION_GROUP

    @classmethod
    def __load_permission_class(cls):
        """载入可用的权限类别
        """
        mp = {}

        classes = list(cls.mongo[cls.PERMISSION_CLASS_TABLE].find())

        if not classes:
            classes = cls.gen_permission_class()

        for c in classes:
            mp[int(c["key"])] = c

        cls.CURRENT_PERMISSION_CLASS = mp

        return cls.CURRENT_PERMISSION_CLASS

    @classmethod
    def get_permission_group(cls, gid=None):
        """获取一个或所有角色组"""
        if not cls.CURRENT_PERMISSION_GROUP:
            cls.__load_permission_group()

        if gid is None:
            return cls.CURRENT_PERMISSION_GROUP or {}
        else:
            return cls.CURRENT_PERMISSION_GROUP["group"][int(gid)]

    @classmethod
    def get_permission_class(cls, class_key=None):
        """
        获取一个或所有权限类别
        :param class_key:
        :return:
        """
        if not cls.CURRENT_PERMISSION_CLASS:
            cls.__load_permission_class()

        if class_key is None:
            return cls.CURRENT_PERMISSION_CLASS
        else:
            return cls.CURRENT_PERMISSION_CLASS[int(class_key)]

    @classmethod
    def create_group(cls, **group):
        """
        创建一个新的角色组
        :param group: 提供配置文件
        :param parent_gid:   指定父角色，在其后面添加下级角色
        :return:
        """

        parent_pid = group.pop("parent_pid", None)
        if parent_pid == 0:
            return False

        max_gid = 0
        max_pid = 0
        for g in cls.mongo[cls.PERMISSION_TABLE].find():
            if g["gid"] > max_gid:
                max_gid = g["gid"]
            if g["pid"] > max_pid:
                max_pid = g["pid"]

        new_gid = max_gid + 1
        new_pid = max_pid + 1

        if parent_pid is not None:
            new_pid = parent_pid + 1
            cls.mongo[cls.PERMISSION_TABLE].update({"pid": {"$gte": parent_pid + 1}}, {"$inc": {"pid": 1}}, multi=True)

        group["_id"] = str(ObjectId())
        group["available"] = True
        group["gid"] = new_gid  # 权限组ID，大小与父子关系没有必然联系
        group["pid"] = new_pid  # 父子关系，1 > 2 > 3，值越小层级越高
        group["permission_class"] = group.get("permission_class", {})
        group["create_at"] = int(time.time())

        if not cls.check_permission_group(group):
            raise Exception("group config is error.")

        cls.mongo[cls.PERMISSION_TABLE].insert(group)
        cls.__load_permission_group()

        return group

    @classmethod
    def copy_group(cls, copy_gid):
        """
        复制指定的角色组并创建
        :param copy_gid:
        :return:
        """
        copy_group = cls.CURRENT_PERMISSION_GROUP["group"][copy_gid]
        new_group = cls.create_group(**copy_group)
        cls.__load_permission_group()
        return new_group

    @classmethod
    def add_class2group(cls, gid, **kwargs):
        """
        给某个角色组加上一个权限类别
        :param kwargs: {"1": 23}
        :param gid:
        :return:
        """
        s = {}
        for class_key, mask in kwargs.items():
            s["permission_class.%s" % int(class_key)] = int(mask)

        cls.mongo[cls.PERMISSION_TABLE].update({"gid": gid}, {"$set": s})
        cls.__load_permission_group()
        cls.__load_permission_class()

    @classmethod
    def create_class(cls, _class=None, nums=MAX_NUMS):
        """
        创建新的权限类
        :param _class:  权限类对象
        :param nums:  bit 权限位数
        :return:
        """
        if _class and not cls.check_permission_class(_class):
            return False

        if _class and cls.mongo[cls.PERMISSION_CLASS_TABLE].find_one({"key": int(_class["key"])}):
            return False

        if not _class:
            key = cls.mongo[cls.PERMISSION_CLASS_TABLE].find().sort("key", -1)[0]["key"]
            _class = cls.gen_class_obj(key + 1, nums)

        _class.setdefault("_id", str(ObjectId()))
        cls.mongo[cls.PERMISSION_CLASS_TABLE].insert(_class)

        _class.pop("_id", None)
        cls.__load_permission_class()
        return _class

    @classmethod
    def create_bit(cls, class_key, name):
        """
        生成新的权限位
        :param class_key:
        :param name:
        :return:
        """
        if cls.CURRENT_PERMISSION_CLASS is None:
            raise Exception("permission class is None, you must load it.")

        permission_class = cls.CURRENT_PERMISSION_CLASS[int(class_key)]

        bit_mp = permission_class["bit_mp"]

        if len(bit_mp) >= cls.MAX_NUMS:
            raise Exception("permission class bit nums must <= %d." % cls.MAX_NUMS)

        max_bit = 0
        max_sort = 0
        for k, item in bit_mp.items():
            if int(k) > max_bit:
                max_bit = int(k)
            if item["sort"] > max_sort:
                max_sort = item["sort"]

        max_bit += 1
        max_sort += 1
        if max_bit >= cls.MAX_NUMS:
            raise Exception("permission class bit nums must <= %d." % cls.MAX_NUMS)

        new_bit = cls.bit_mp_obj(name, max_sort)
        cls.mongo[cls.PERMISSION_CLASS_TABLE].update({"key": int(class_key)},
                                                  {"$set": {"bit_mp.%d" % max_bit: new_bit}})
        cls.__load_permission_class()

        return new_bit

    @classmethod
    def update_group_name(cls, gid, new_name):
        """
        更新用户组名称，但是特性不能改，如果与原有特性不一样，需要配合修改以用于判断的代码
        TODO 原则上不推荐修改
        :param gid:
        :param new_name:
        :return:
        """
        if gid not in cls.CURRENT_PERMISSION_GROUP["group"]:
            raise errors.ErrorInvalid("不存在该角色")

        for g in cls.CURRENT_PERMISSION_GROUP["group"].values():
            if g["name"].encode("u8") == new_name:
                return

        cls.mongo[cls.PERMISSION_TABLE].update({"gid": int(gid)}, {"$set": {"name": new_name}})
        return cls.__load_permission_group()

    @classmethod
    def update_group_available(cls, gid, available):
        """
        更新用户组名称
        TODO 原则上不推荐修改
        :param gid:
        :param new_name:
        :return:
        """
        if type(available) is not bool:
            return

        cls.mongo[cls.PERMISSION_TABLE].update({"gid": int(gid)}, {"$set": {"available": available}})
        cls.__load_permission_group()

    @classmethod
    def update_class_name(cls, class_key, new_name):
        """修改权限类名
        TODO 原则上不推荐修改
        """
        if class_key not in cls.CURRENT_PERMISSION_CLASS[class_key]:
            raise errors.ErrorInvalid("不存在该权限类")

        if cls.mongo[cls.PERMISSION_CLASS_TABLE].find_one({"name": new_name}):
            return False

        cls.mongo[cls.PERMISSION_CLASS_TABLE].update({"key": int(class_key)}, {"$set": {"name": new_name}})
        return cls.__load_permission_class()

    @classmethod
    def update_class_available(cls, class_key, available):
        """修改权限类可用
        TODO 原则上不推荐修改
        """
        cls.mongo[cls.PERMISSION_CLASS_TABLE].update({"key": int(class_key)}, {"$set": {"available": bool(available)}})
        return cls.__load_permission_class()

    @classmethod
    def update_bit_name(cls, class_key, bit, new_name):
        """修改单个权限位的权限名
        注意：权限如果被修改了，所有使用过该位的地方都要跟着改
        TODO 原则上不推荐修改
        """
        if class_key not in cls.CURRENT_PERMISSION_CLASS:
            raise errors.ErrorInvalid("不存在该权限类")
        if str(bit) not in cls.CURRENT_PERMISSION_CLASS[class_key]["bit_mp"]:
            raise errors.ErrorInvalid("不存在该权限")

        for c in cls.mongo[cls.PERMISSION_CLASS_TABLE].find():
            for b, name in c["bit_mp"].items():
                if name["name"].encode("u8") == new_name:
                    return

        cls.mongo[cls.PERMISSION_CLASS_TABLE].update({"key": int(class_key)},
                                                  {"$set": {"bit_mp.%s.name" % str(bit): new_name}})
        return cls.__load_permission_class()

    @classmethod
    def update_batch_bit_name(cls, class_key, **kwargs):
        """
        修改多个权限位的权限名,支持同时修改多个权限位
        注意：权限如果被修改了，所有使用过该位的地方都要跟着改
        TODO 原则上不推荐修改
        :param class_key:
        :param kwargs: {"0": {"available": True, "name": "新权限名"},
                        "1": {"name": "新权限名"},
                        "2": {"available": True}
                    }
        :return:
        """
        s = {}

        for k, v in kwargs.items():
            if not v:
                continue
            if "available" in v:
                s["bit_mp.%s.available" % str(k)] = v["available"]

            if "name" in v and v["name"]:
                s["bit_mp.%s.name" % str(k)] = v["name"]

        if s:
            cls.mongo[cls.PERMISSION_CLASS_TABLE].update({"key": int(class_key)}, {"$set": s})
            return cls.__load_permission_class()

    @classmethod
    def update_bit_available(cls, class_key, bit, available):
        """修改权限位可用
        TODO 原则上不推荐修改
        """
        cls.mongo[cls.PERMISSION_CLASS_TABLE].update({"key": int(class_key)},
                                                  {"$set": {"bit_mp.%s.available" % str(bit): bool(available)}})
        return cls.__load_permission_class()

    @classmethod
    def update_permission(cls, gid, class_key, **bits):
        """
        按单个角色组修改权限位值
        :param gid:  角色组ID
        :param class_key: 权限组类
        :param bits: 权限位，{"0": True/1, "1": False/0, "54": True/1}
        :return:
        """
        if int(gid) not in cls.CURRENT_PERMISSION_GROUP["group"]:
            raise errors.ErrorInvalid("不存在该角色")

        ts = int(time.time())
        # cls.mongo[cls.PERMISSION_TABLE].update({"gid": gid}, {"$bit": {"i": {"xor": 1}}})
        mask = cls.mongo[cls.PERMISSION_TABLE].find_one({"gid": int(gid)})["permission_class"][str(class_key)]
        new_mask = mask

        for k, v in bits.items():
            if type(v) in [None, bool]:
                v = v == True
            else:
                v = int(v) if v.isdigit() else v == True

            if not mask & 1 << int(k) and v:  # 勾选权限
                new_mask += 1 << int(k)
            elif mask & 1 << int(k) and not v:  # 取消权限
                new_mask -= 1 << int(k)

        if new_mask != mask:
            cls.mongo[cls.PERMISSION_TABLE].update({"gid": int(gid)}, {
                "$set": {"permission_class.%s" % str(class_key): new_mask, "update_at": ts}})
            cls.reload()
        return new_mask

    @classmethod
    def update_batch_permission(cls, **kwargs):
        """
        按多个角色组修改权限位值
        :param kwargs: {"1": {"1": {"0": 1, "1": 1}, "2": ..}, "2": {..}}
        :param gid:  角色组ID
        :param class_key: 权限组类
        :param bits: 权限位，{"0": True/1, "1": False/0, "54": True/1}
        :return:
        """
        for gid, permission_class_mp in kwargs.items():
            for class_key, bits in permission_class_mp.items():
                cls.update_permission(gid, class_key, **bits)
        cls.reload()
        return cls.CURRENT_PERMISSION_GROUP["group"]

    @classmethod
    def get_parent_group(cls, gid):
        """获取上级角色组，"""
        parent_group = {}
        me_pid = cls.CURRENT_PERMISSION_GROUP["group"][int(gid)]["pid"]
        for pid, g in cls.CURRENT_PERMISSION_GROUP["parent"].items():
            # 角色组可以被删除，所以有可能上级不是pid-1,而是pid-2,
            if pid >= me_pid:
                continue
            if g["available"] is not True:
                continue

            parent_group = g

        return parent_group

    @classmethod
    def get_parent_groups(cls, gid):
        """获取所有直系父级、父父级、叔父级"""
        parent_groups = []
        parent_group = cls.get_parent_group(gid)

        if parent_group:
            parent_groups.append(parent_group)

            # 叔父
            brother_groups = cls.get_brother_groups(parent_group["gid"])
            if brother_groups:
                parent_groups += brother_groups

            parent_gid = str(parent_group.get("pid"))
            if str(parent_gid) in cls.CURRENT_PERMISSION_GROUP["group"] and str(parent_gid) != str(parent_group["gid"]):
                parent_groups.extend(cls.get_parent_groups(parent_group["gid"]))

        return cls.parse_ppid_groups(*parent_groups)

    @classmethod
    def get_brother_groups(cls, gid):
        """
        获取gid的兄弟角色
        :param gid:
        :return:
        """
        brother_groups = []
        group = cls.CURRENT_PERMISSION_GROUP["group"][int(gid)]
        parent_gid = group.get("pid")
        if parent_gid == group["gid"]:
            return brother_groups

        for gid, item in cls.CURRENT_PERMISSION_GROUP["group"].items():
            if gid == group["gid"]:
                continue
            if str(item["gid"]) != str(group["gid"]) and item.get("pid") == parent_gid and item["available"] == 1:
                brother_groups.append(item)

        return brother_groups

    @classmethod
    def get_child_group(cls, gid):
        """获取下级角色组"""
        parent_group = {}
        me_pid = cls.CURRENT_PERMISSION_GROUP["group"][int(gid)]["pid"]
        for pid, g in cls.CURRENT_PERMISSION_GROUP["parent"].items():
            # 角色组可以被删除，所以有可能上级不是pid+1,而是pid+2
            if pid <= me_pid:
                continue
            if g["available"] is not True:
                continue

            parent_group = g

        return parent_group

    @classmethod
    def get_child_groups(cls, gid):
        """获取直系子孙级"""
        child_groups = []
        child_group_list = cls.get_child_group(gid)
        child_groups.extend(child_group_list)
        for child in child_group_list:
            child_groups.extend(cls.get_child_groups(child["gid"]))

        s = []
        for child in child_groups:
            if child["available"] != 1:
                continue
            s.append(child)
        return s

    @classmethod
    def get_group_name(cls, gid):
        """用户组->名称"""
        group_permission = cls.CURRENT_PERMISSION_GROUP["group"][int(gid)]
        return group_permission["name"]

    @classmethod
    def is_parent_group(cls, gid, parent_gid):
        """
        判断是否是上级角色
        :param parent_gid:
        :return:
        """
        step = 1
        group_mp = cls.CURRENT_PERMISSION_GROUP["group"]
        parent_mp = cls.CURRENT_PERMISSION_GROUP["parent"]

        pid = group_mp[int(gid)]["pid"]
        parent_pid = group_mp[parent_gid]["pid"]

        if pid <= parent_gid:
            return False

        keys = parent_mp.keys()
        keys.sort(reverse=True)
        for k in keys:
            if k >= pid:
                continue

            group = parent_mp[k]

            # 相邻多个上级被禁用，不算
            if group["available"] == False:
                continue
            else:
                if group["pid"] != parent_pid:
                    step += 1
                else:
                    break

        return step == 1

    @classmethod
    def is_child_group(cls, gid, child_gid):
        """
        判断是否是下级角色
        :param parent_gid:
        :return:
        """

        step = 1
        group_mp = cls.CURRENT_PERMISSION_GROUP["group"]
        child_mp = cls.CURRENT_PERMISSION_GROUP["parent"]

        pid = group_mp[int(gid)]["pid"]
        child_gid = group_mp[str(child_gid)]["pid"]

        if pid >= child_gid:
            return False

        keys = child_mp.keys()
        keys.sort()
        for k in keys:
            if k <= pid:
                continue

            group = child_mp[k]

            # 相邻多个下级被禁用，不算
            if group["available"] == False:
                continue
            else:
                if group["pid"] != child_gid:
                    step += 1
                else:
                    break

        return step == 1

    @classmethod
    def reload(cls):
        """修改权限设置后,更新mp"""
        cls.CURRENT_PERMISSION_GROUP = None
        cls.CURRENT_PERMISSION_CLASS = None
        cls.__load_permission_group()
        cls.__load_permission_class()
        cls.permission_changed()

    @classmethod
    def check_permission(cls, gid, class_key, permission):
        """检测用户所在的组 是否有某个权限
        :param mask: 当前的权限设定
        :param permission: 需要判断是否有这个权限 xxxxx_PERMISSION 2的n阶数 多个不同权限可以相加, 检测是否同时拥有多个权限
        ANONYMOUS_PERMISSION = 0
        """

        if cls.IS_DEBUG:
            return True

        # 非必要,为了效率
        if permission == 0:
            return True

        group_permission = cls.get_permission_group(gid)

        if group_permission["available"] is False:
            return True

        mask = group_permission["permission_class"].get(str(class_key), 0)

        return mask & permission == permission

    @classmethod
    def crontab_update(cls):
        """检测是否需要更新"""
        last_version = int(cls.redis.get(PERMISSION_VERSION_CACHE_KEY) or -1)
        if cls.__VERSION__ == last_version:
            return False

        logging.debug("crontab_update: true")
        cls.reload()
        cls.__VERSION__ = last_version
        return True

    @classmethod
    def permission_changed(cls):
        pass
        # cls.__VERSION__ = int(cls.redis.incrby(PERMISSION_VERSION_CACHE_KEY, 1))

    @classmethod
    def test(cls):
        print("======所有角色组=======")
        print(cls.get_permission_group())
        print
        print("======角色组1=======")
        print(cls.get_permission_group(1))
        print
        print("======所有权限类别=======")
        print(cls.get_permission_class())





#!/usr/bin/env python
# -*- coding:utf-8 -*-
import os
from lcyframe import route
from base import BaseHandler, helper

@route("/imagecode")
class ImageCodeHandler(BaseHandler):
    """
    图片验证码
    """
    @helper.admin(0)
    def get(self):
        """
        获取验证码


        Request extra query params:
        - uuid # 随机串 type: str


        :return:
        :rtype:
        """
        code, code_data = self.model.BasicsModel.send_imagecode(self.params["uuid"])
        self.write_success(data={
            "code_data": code_data,
            "uuid": self.params["uuid"]
        })

    @helper.admin(0)
    def post(self):
        """
        校验图形验证码
        
        
        Request extra query params:
        - stype # 类型 type: int
        - uuid # 随机串 type: str
        - imgcode # 图形验证码 type: str
        

        :return:
        :rtype:
        """
        
        code = self.params["code"]
        uuid = self.params["uuid"]
        if not self.model.BasicsModel.vaild_imagecode(uuid, code):
            raise self.api_error.ErrorCommon("验证码错误")
        self.write_success()



@route("/smscode")
class SMSCodeHandler(BaseHandler):
    """
    短信验证码
    """
    @helper.admin(0)
    def get(self):
        """
        获取短信验证码
        
        
        Request extra query params:
        - stype # 类型,预留 type: int
        - uuid # 随机串,预留 type: str
        - mobile # 随机串 type: str
        

        :return:
        :rtype:
        """
        sms_code = self.model.BasicsModel.send_smscode(
            self.params["uuid"],
            self.params["mobile"],
            self.params["stype"]
        )
        self.write_success(data={"code": sms_code, "uuid": self.params["uuid"]})

    @helper.admin(0)
    def post(self):
        """
        短信验证码


        Request extra query params:
        - stype # 类型,预留 type: int
        - uuid # 随机串,预留 type: str
        - mobile # 随机串 type: str
        - smscode # 验证码 type: int


        :return:
        :rtype:
        """

        code = self.params["code"]
        uuid = self.params["uuid"]
        if not self.model.BasicsModel.vaild_smscode(uuid, code):
            raise self.api_error.ErrorCommon("验证码错误")
        self.write_success()

@route("/sendemail")
class SendEmailHandler(BaseHandler):
    @helper.admin(0)
    def get(self):
        """
        测试发送邮件
        """

        """
        发送试用申请通过通知
        :param company: 试用申请企业名,
        :param username: 用户名称
        :param account: 用户账号
        :param email: 接收邮箱
        :param password: 密码
        :param domain_name: 授权域名
        :param activate_url: 激活链接
        """
        raise
        args = ["商户", "用户名称", "用户账号", ["123220663@qq.com"], "密码", "授权域名", "激活链接"]
        self.model.ActivateModel.email_avtivate_link(*args,
                                                     license="eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJleHAiOjE2OTM2MTI4MDAsImlhdCI6MTY2MjExNDMyMywiZGF0YSI6eyJhdXRob3JpemF0aW9uIjo2OSwiY29tcGFueSI6MjcsInNlcnZpY2VfbGlmZSI6MzAsInByb2R1Y3RfaWQiOjcsInByb2R1Y3RfbmFtZSI6Ilx1NjgwN1x1NTFjNlx1NGVhN1x1NTRjMSIsInByb2R1Y3RfdHlwZSI6MSwiYXV0aF9zdGF0ZSI6MCwidmVyc2lvbiI6MS4wLCJhY3RpdmVfbW9kZSI6MCwiaXNzdWVkX2F0IjoiMjAyMi0wOS0wMiAxMDoyNToyMyIsImF1dGhfdmFsaWRfYXQiOiIyMDIzLTA5LTAyIDAwOjAwOjAwIn19.D0NnyvK7323o50bEFW5PPuTX4Pz7oFw80cN1Do_2Bw8",
                                                     tryapply=True
                                                     )

        # 发送邮件激活
        # args = ["商户", ["123220663@qq.com"]]
        # self.model.ActivateModel.email_avtivate_link(*args, license="eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJleHAiOjE2OTM2MTI4MDAsImlhdCI6MTY2MjExNDMyMywiZGF0YSI6eyJhdXRob3JpemF0aW9uIjo2OSwiY29tcGFueSI6MjcsInNlcnZpY2VfbGlmZSI6MzAsInByb2R1Y3RfaWQiOjcsInByb2R1Y3RfbmFtZSI6Ilx1NjgwN1x1NTFjNlx1NGVhN1x1NTRjMSIsInByb2R1Y3RfdHlwZSI6MSwiYXV0aF9zdGF0ZSI6MCwidmVyc2lvbiI6MS4wLCJhY3RpdmVfbW9kZSI6MCwiaXNzdWVkX2F0IjoiMjAyMi0wOS0wMiAxMDoyNToyMyIsImF1dGhfdmFsaWRfYXQiOiIyMDIzLTA5LTAyIDAwOjAwOjAwIn19.D0NnyvK7323o50bEFW5PPuTX4Pz7oFw80cN1Do_2Bw8")
        return self.write_success()



@route("/download")
class DownloadHandler(BaseHandler):
    @helper.admin(0)
    def get(self):
        """
        下载文件
        stype: 1授权文件

        """
        filepath = {
            1: self.app_config["data_config"]["authorizedoc"],
            2: self.app_config["data_config"]["followup"]
        }[self.params["stype"]]
        file_id = self.params["name"]
        path = os.path.join(self.app_config["data_config"]["data_base"], filepath, file_id)
        if not os.path.exists(path):
            raise self.api_error.ErrorCommon("不数据不存在")

        with open(os.path.join(self.app_config["data_config"]["data_base"], filepath, file_id), "rb") as p:
            # while True:
            data = p.read()

            self.set_header("charset", 'UTF-8')
            # self.set_header("content_type", 'application/download')
            self.set_header("Content_Type", 'application/octet-stream')
            self.set_header("Content_Disposition", f'attachment; filename="{file_id}"')
            # super(BaseHandler, self).write(base64.b64encode(data).decode())
            super(BaseHandler, self).write(data)



#!/usr/bin/env python
# -*- coding:utf-8 -*-


"""
redis Keys
"""
TOKEN_MARK = "token:%s"    # 登陆标记
NONCE_MARK = "nonce:%s"    # 重复请求标记

imagecode_key = "imagecode:%s"
imagecode_exp = 1 * 60

# 短信验证码每个账号每天最多只能收到10条
# 短信验证码每个IP每天最多只能收到500条
sms_times = 10
smscode_interval = "smsinterval:%s"     # 间隔1分钟
smscode_times = "smstimes:%s"           # 每天发送最多10条
smscode_key = "smscode:%s"              # 验证码
smscode_exp = 5 * 60                    # 有效期




#!/usr/bin/env python
# -*- coding:utf-8 -*-
import webbrowser, os
from ..base import BaseHandler
import platform

from tornado.web import HTTPError, StaticFileHandler

class StaticHandler(BaseHandler):
    """
    static
    """
    def get(self, path):
        if not path:
            raise HTTPError(404)
        else:
            return StaticFileHandler.get(path)


# coding=utf-8
import os
import logging
from lcyframe.libs import utils

class MqRoute(object):
    """消息队列路由"""
    _workers = {}

    def __init__(self, event=None):
        self.event = event

    def __call__(self, _handler):
        event = self.event and self.event or _handler.__name__
        _handler.from_route = True
        self._workers[event] = _handler
        return _handler

    @classmethod
    def tasks(cls, _handler):
        for event_name, event_obj in _handler.__dict__.items():
            if not event_obj:
                continue
            if type(event_obj) == str:
                continue
            if event_name in cls._workers:
                raise Exception("The event is exists multiple: `%s`" % event_name)
            cls._workers[event_name] = getattr(_handler, event_name)
        return cls._workers

    @classmethod
    def get_workers(cls, ROOT, workers_path):
        if not ROOT:
            raise Exception("the project dir path must been give， and None.")

        if not isinstance(workers_path, list):
            workers_path = [workers_path]

        if workers_path is None:
            raise Exception("args workers_path is not allow empty")

        if not cls._workers:
            for work in workers_path:
                for root, dirs, files in os.walk(utils.fix_path(os.path.join(ROOT, work))):
                    for file in files:
                        if file.startswith("__"):
                            continue
                        if file.endswith(".pyc"):
                            continue
                        if not file.endswith(".py"):
                            continue
                        model_name = root.replace(ROOT, "").lstrip("/").replace("/", ".") + "." + file.rstrip(".py")
                        __import__(model_name, globals(), locals(), [model_name], 0)
                        logging.debug("register mq workers [%s.py] success!" % model_name)
        return cls._workers


#!/usr/bin/env python
# -*- coding:utf-8 -*-

from Crypto.Cipher import AES as _AES
from binascii import b2a_hex, a2b_hex
from base64 import b64decode, b64encode
import json

class AES():
    def __init__(self, key=None,
                 iv=b'3eddc41dd41239c8',
                 mode=_AES.MODE_ECB,
                 padding="zeropadding",
                 print_model="hex"):
        # 这里密钥key 长度必须为16（AES-128）, 24（AES-192）,或者32 （AES-256）Bytes 长度
        self.key = key if key else "ODQJmzxHP83YTV5aysc90bUMon1GqhKF" # "054eda8889ea6240"
        self.mode = mode                    # 模式
        self.iv = iv                        # 偏移量
        self.padding = padding              # 填充方式： 默认 zeropadding： 0补全；pkcs7padding；pkcs5padding；等
        self.print_model = print_model      # 输出格式 hex：十六进制 base64：base64格式

    def encode(self, text):
        if not isinstance(text, str):
            text = json.dumps(text)
        try:
            cryptor = _AES.new(self.key.encode('utf-8'), self.mode, self.iv)
        except:
            cryptor = _AES.new(self.key.encode('utf-8'), self.mode)
        length = 16     # 如果text不足16位就用空格补足为16位， 如果大于16当时不是16的倍数，那就补足为16的倍数
        count = len(text)
        if count < length:
            add = (length - count)
            # \0 backspace
            text = text + ('\0' * add)
        elif count > length and count % length != 0:
            add = (length - (count % length))
            text = text + ('\0' * add)
        self.ciphertext = cryptor.encrypt(text.encode('utf-8'))
        # 因为AES加密时候得到的字符串不一定是ascii字符集的，输出到终端或者保存时候可能存在问题
        # 所以这里统一把加密后的字符串格式化输出
        if self.print_model == "hex":
            return b2a_hex(self.ciphertext).decode()
        else:
            return b64encode(self.ciphertext).decode()

    # 解密后，去掉补足的空格用strip() 去掉
    def decode(self, text):
        text = text if type(text) == bytes else text.encode('utf-8')
        try:
            cryptor = _AES.new(self.key.encode('utf-8'), self.mode, self.iv)
        except:
            cryptor = _AES.new(self.key.encode('utf-8'), self.mode)
        if self.print_model == "hex":
            plain_text = cryptor.decrypt(a2b_hex(text)).decode()
        else:
            plain_text = cryptor.decrypt(b64decode(text)).decode()

        try:
            return json.loads(plain_text.rstrip('\0'))
        except ValueError:
            return plain_text.rstrip('\0')

if __name__ == '__main__':
    pc = AES("debe8e44b8b68d3a528da563e147e7bb")  #  054eda8889ea6240 初始化密钥
    s = "a:b"
    d = pc.encode(s)
    print("加密:", d)

    d = pc.decode(d)  # 解密
    print("解密:", d)


# -*- coding:utf-8 -*-
import base64
from Crypto.Cipher import AES
from Crypto import Random
from Crypto.Hash import MD5
from lcyframe.libs import utils
from lcyframe.libs.aes import AES as sAES

BS = AES.block_size
pad = lambda s: s + (BS - len(s) % BS) * chr(BS - len(s) % BS)
unpad = lambda s: s[:-ord(s[len(s) - 1:])]


def openssl_evp_byte2key(password, salt, key_len, iv_len):
    """
    Derive the key and the IV from the given password and salt.

    see OpenSSL man:
    https://www.openssl.org/docs/crypto/EVP_BytesToKey.html
    """
    # from hashlib import md5
    # dtot = md5(password + salt).digest()
    # d = [ dtot ]
    # while len(dtot)<(iv_len+key_len):
    #     d.append( md5(d[-1] + password + salt).digest() )
    #     dtot += d[-1]
    # return dtot[:key_len], dtot[key_len:key_len+iv_len]
    dtot = MD5.new('{0}{1}'.format(password, salt)).digest()
    d = [dtot]
    while len(dtot) < (iv_len + key_len):
        d.append(MD5.new('{0}{1}{2}'.format(d[-1], password, salt)).digest())
        dtot += d[-1]
    return dtot[:key_len], dtot[key_len:key_len+iv_len]


def openssl_aes_decrypt(encoded_text, password_phase):
    """
    decode an openssl-compatible 256-bit AES cipher text
    :param encoded_text:
    :param password_phase:
    :return:
    """
    encrypted = base64.b64decode(encoded_text)
    # salt = encrypted[8:16]
    salt = encrypted[0:8]
    data = encrypted[8:]
    key, iv = openssl_evp_byte2key(password_phase, salt, 32, 16)
    dec = AES.new(key, AES.MODE_CBC, iv)
    plain_text = dec.decrypt(data)
    return unpad(plain_text)


def openssl_aes_encrypt(plain_text, password_phase):
    salt = Random.new().read(8)
    key, iv = openssl_evp_byte2key(password_phase, salt, 32, 16)
    enc = AES.new(key, AES.MODE_CBC, iv)
    plain_text = pad(plain_text)
    return base64.b64encode(salt + enc.encrypt(plain_text))

def salt_aes_decrypt(encoded_text, password_phase):
    """
    > * 1：随机生成30位由字母与数字组成的字符串，并转为小写，得到salt。
    > * 2：将Payload消息体，按照原来机器人的AES加密方法进行加密，得到plain_text
    > * 3：将salt + plain_text，得到最终密文。
    """

    salt = encoded_text[0:30]
    data = encoded_text[30:]
    plain_text = sAES(password_phase).decode(data.decode())

    return plain_text


def salt_aes_encrypt(plain_text, password_phase):
    """
    > * 1：将密文从第31位(含)开始，截取至末尾，得到字符串plain_text。
    > * 2：将plain_text消息体，按照原来机器人的AES加密方法进行解密，得到明文Payload。
    :param plain_text:
    :param password_phase:
    :return:
    """
    enc = sAES(password_phase).encode(plain_text)
    salt = utils.gen_random_sint(30).lower()

    return salt + enc


def encrypt(plain_text, password_phase, mod=1):
    if int(mod) == 1:
        return salt_aes_encrypt(plain_text, password_phase)
    else:
        return openssl_aes_encrypt(plain_text, password_phase)

def decrypt(plain_text, password_phase, mod=1):
    if int(mod) == 1:
        return salt_aes_decrypt(plain_text, password_phase)
    else:
        return openssl_aes_decrypt(plain_text, password_phase)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import sys
import json
import logging
import random
import time
import types
from functools import wraps
from traceback import format_exc

from bson.objectid import ObjectId
from tornado.web import HTTPError
from tornado.web import RequestHandler
from lcyframe.libs import errors
from lcyframe.libs import utils
from lcyframe.libs.funts import get_return


class BaseHandler(RequestHandler):
    __structure__ = {}
    logging = logging
    app_config = {}
    Permission = None
    model = None
    mongo = None
    mysql = None
    redis = None
    ssdb = None
    mq = None
    nsq = None
    mqtt = None
    celery = None
    beanstalk = None
    qiniu = None
    emchat = None
    aes = None
    graylog = None
    sms = None
    smtp = None

    def __init__(self, application, request, **kwargs):
        self.application = application
        self.request = request
        super(BaseHandler, self).__init__(application, request, **kwargs)
        if request.headers.get("cdn-src-ip", None):
            request.remote_ip = request.headers["cdn-src-ip"]
        elif request.headers.get("X-Forwarded-For", None):
            try:
                request.remote_ip = request.headers["X-Forwarded-For"].split(",")[0]
            except Exception as e:
                request.remote_ip = self.request.headers.get('X-Real-Ip', '')
        self.cors_config = self.application.app_config.get("cors_config", {})

    def prepare(self):
        # super(BaseHandler, self).prepare()
        #     self.permission.has_permission(self.request)
        if self.request.method == "OPTIONS" and self._allow_cors:
            self.set_cors_header()
            self.set_status(204)
            self.finish()
            return

    @property
    def _allow_cors(self):
        """
        校验授信跨域来路
        :return:
        """
        self.cors_config = self.application.app_config.get("cors_config", {})
        allow = self.cors_config and self.cors_config.get("allow", False)
        self.origin = self.request.headers.get("origin")
        self.cors_config["origin"].append(self.application.app_config["wsgi"]["host"])
        auto_origin = self.cors_config["origin"]
        if not allow or ("*" not in auto_origin and self.origin not in auto_origin):
            raise errors.ErrorCorsUri
        return True

    def set_cors_header(self):
        """通过授信校验后，允许设置响应头"""
        self.set_header("Content-Type", self.cors_config.get("content-type", "text/html") + "; charset=utf-8")
        self.set_header("Access-Control-Allow-Origin", "*" if "*" in self.cors_config.get("origin", "origin") else self.request.headers.get("origin", ""))
        self.set_header("Access-Control-Allow-Headers", ",".join(self.cors_config.get("headers", ["*"])))
        self.set_header("Access-Control-Allow-Credentials", "true")
        self.set_header("Access-Control-Allow-Methods", ",".join(self.cors_config.get("methods", ["GET", "POST", "PUT", "DELETE"])))
        self.set_header("Access-Control-Max-Age", self.cors_config.get("max-age", 60))

    def write_error(self, status_code, **kwargs):
        self.cors_config = self.application.app_config.get("cors_config", {})
        allow = self.cors_config and self.cors_config.get("allow", False)
        if allow:
            self.set_cors_header()

        if status_code == 404:
            return self.write_failed(code=status_code, msg="Not Found", name="HTTP 404")
        elif status_code == 405:
            return self.write_failed(code=status_code, msg="HTTP 405: Method/URL Not Allowed", name="HTTP 405")
        elif status_code == 500:
            # typ, value, tb = sys.exc_info()
            # logging.error("Uncaught exception %s\n%r", self._request_summary(), self.request, exc_info=(typ, value, tb))

            error = kwargs["exc_info"][1]
            # 把除跨域异常外所有用raise抛出的内置异常类型，都转为200返回，内部保留500异常log
            if isinstance(error, errors.ApiError) and error.code != errors.ErrorCorsUri.code:
                self.set_status(200, "HTTP STATUS 200")

            logging.error("异常时间：" + self.utils.int_to_date_string(self.utils.now(), "%Y-%m-%d %H:%M:%S"))
            if not hasattr(error, "message"):
                message = "%s:%s" % (type(error).__name__, error.args[0]) if error.args else "未知错误"
                logging.error("异常信息：" + message)
            else:
                logging.error("异常码：" + str(error.code))
                logging.error("异常信息：" + error.message)

            logging.error("请求地址: %s %.2fms", self.request.uri, 1000.0 * self.request.request_time())
            logging.error("请求参数: " + (json.dumps(utils.check2json(utils.pparams(self.request, params=self.source_params))) if hasattr(self, "source_params") else str(utils.pparams(self.request) or {})))
            if not hasattr(error, "code_name") or not hasattr(error, "code"):
                return self.write_failed()
            else:
                return self.write_failed(code=error.code, msg=error.message, name=error.code_name)
        else:
            return self.write_failed()

    def write_success(self, data=None):
        request_time = 1000.0 * self.request.request_time()
        is_slow_request = request_time > int(self.application.app_config["wsgi"].get("slow_request_time", 1000))
        p = self.source_params if hasattr(self, "source_params") else None
        if is_slow_request:
            logging.warning("[%s] %d %s %.2fms", "慢请求", self.get_status(), self._request_summary(), request_time)

        if hasattr(self, "graylog") and self.graylog:
            if self.application.app_config["graylog_config"].get("mode", 0) or is_slow_request:
                self.graylog.from_server(short_message="OK", code=0, host=self.request.host, request_time=int(request_time), remote_ip=self.request.remote_ip,
                                         uri=self.request.path, params=utils.pparams(self.request, params=p), method=self.request.method)
        return self.write({"code": 0, "msg": "OK", "data": get_return(self, data or {})})

    def write_failed(self, code=1, msg="Error. unknown error", name="HTTP 500"):
        if hasattr(self, "graylog") and self.graylog:
            p = self.params if hasattr(self, "params") else {}
            self.graylog.from_server(short_message=msg, code=code, track=format_exc()[-500:], host=self.request.host, header=self.request.headers._dict,
                                     remote_ip=self.request.remote_ip, uri=self.request.path, params=utils.pparams(self.request, params=p), method=self.request.method)
        if name:
            self.write({"code": code, "msg": msg, "name": name, "data": None})
            return
        else:
            self.write({"code": code, "msg": msg, "data": None})
            return

    def write_html(self, template_name, **kwargs):
        return self.render(template_name, **kwargs)

    def write(self, chunk):
        # from bson import json_util
        # chunk = json_util.dumps(chunk)
        chunk = utils.check2json(chunk)
        if isinstance(chunk, (dict, list, tuple)):
            chunk = json.dumps(chunk, ensure_ascii=False)
        return super(BaseHandler, self).write(chunk)

    # def _request_summary(self):
    #     return "- [%s] %s %s (%s)" % (Copyright.frame_name, self.request.method, self.request.uri,
    #                            self.request.remote_ip)

    def json_callback(self, data):
        """json回调跨域
        """
        jsoncallback = self.get_argument("jsoncallback", None)
        if jsoncallback:
            self.write("%s(%s)" % (jsoncallback, json.dumps(data)))
        else:
            self.write(data)

    def restrict(self, count, duration=60):
        """接口访问限制"""

        def wrap(method):
            @wraps(method)
            def has_role(self, *args, **kwargs):
                if not hasattr(self, "redis"):
                    return method(self, *args, **kwargs)

                # 获取客户端ip
                ip = None
                if self.request.headers.get("cdn-src-ip"):
                    ip = self.request.headers.get("cdn-src-ip")
                elif self.request.headers.get("X-Forwarded-For"):
                    z = self.request.headers["X-Forwarded-For"].split(",")
                    if len(z) > 0:
                        ip = z[0]
                ip = ip or self.request.remote_ip

                n = self.redis.incr("restrict:%s" % ip)

                if n == 1:
                    self.redis.expire("restrict:%s" % ip, duration)
                elif n > count:
                    raise HTTPError(503, "Slow Down")

                return method(self, *args, **kwargs)

            return has_role

        return wrap

    @property
    def utils(self):
        return utils

    def get_headers(self, HEADERS=dict()):
        _headers = {}
        if not HEADERS:
            for k, v in self.request.headers.items():
                _headers[k.lower()] = v

        for k, v in HEADERS.items():
            value = self.request.headers.get(k)
            if value is None:
                # 客户端没有传值, 按照默认值类型赋值
                value = v(value) if isinstance(v, types.FunctionType) else v
            else:
                # 按照默认值的类型 转换参数
                value = utils.TypeConvert.apply(v, value)

            _headers[k] = value

        return _headers

    @property
    def app_config(self):
        return self.application.app_config

    @property
    def model(self):
        return BaseModel.model

    @property
    def handler_name(self):
        return self.__class__.__name__

    # @property
    # def Permission(self):
    #     return Permission.initial()
    #     # 复写该方法，已实现自定义的权限逻辑



class BaseModel(object):
    logging = logging
    utils = utils
    Permission = None
    api_error = None
    app_config = {}
    model = None
    mongo = None
    mysql = None
    redis = None
    ssdb = None
    mq = None
    nsq = None
    mqtt = None
    celery = None
    beanstalk = None
    qiniu = None
    emchat = None
    aes = None
    graylog = None
    sms = None
    smtp = None
    
    @classmethod
    def _parse_data(cls, d, **kwargs):
        """
        组装单条数据
        :return:
        :rtype:
        """
        return_data = {}
        if not d:
            return {}

        for k, v in d.items():
            # if k in ["pass_word", "salt", "permission"] and k not in kwargs.get("kwargs", {}):
            #     continue
            return_data[k] = v

        return return_data


class BaseSchema(object):
    """
    is_shard = True                           # 是否分表
    shard_key = "uid"                         # 分表键
    shard_rule()                              # 分表规则，允许复写
    """
    collection = ""
    is_shard = False
    shard_key = ""

    @classmethod
    def ts(cls):
        return int(time.time())

    @classmethod
    def fields(cls):
        return vars(cls())

    @classmethod
    def shard_rule(cls, shard_key_value):
        """
        :param shard_key_value:
        :return: table_value

        This is default rule by `mod10`
        you can overwriter like this

        :example ::
            def shard_rule(shard_key_value):
                do_your_thing
                ...
        """
        return cls.mod10(shard_key_value)

    @classmethod
    def get_shard_table(cls, sql=None, shard_key_value=None):
        """
        默认分表规则
        可以在从schema中复写该方法
        :param shard_key:
        :return:
        """
        if cls.is_shard:
            assert sql is not None or shard_key_value is not None
            assert cls.shard_key in cls.fields()
            value = shard_key_value if shard_key_value is not None else cls.get_value_with_sql(sql)
            return cls.collection + "_" + str(cls.shard_rule(value))
        else:
            return cls.collection

    @classmethod
    def get_value_with_sql(cls, sql):
        """
        :param sql:
        :return:
        """
        if cls.shard_key not in sql:
            raise KeyError("When shard table, shard key must in sql")
        return sql[cls.shard_key]

    @classmethod
    def mod10(cls, shard_key_value=None):
        """
        取模10
        :param mod:
        :return:
        """
        if type(shard_key_value) != int:
            raise KeyError("When shard table, shard key type must int")
        return shard_key_value % 10

    @classmethod
    def dbo(cls, sql=None, shard_key_value=None):
        return cls.mongo[cls.get_shard_table(sql=sql, shard_key_value=shard_key_value)]

    @classmethod
    def get_by_oid(cls, oid, shard_key_value=None, **kwargs):
        """
        单条记录
        :return:
        :rtype:
        """
        return cls.find_one_by_oid(oid, shard_key_value=None, **kwargs)

    @classmethod
    def get_by_spec(cls, spec, shard_key_value=None, **kwargs):
        """
        条件查询
        :param spec:
        :return:
        """
        return cls.find_one(spec)

    @classmethod
    def create(cls, **kwargs):
        """
        创建
        :return:
        :rtype:
        """
        docs = cls.fields()
        for k, v in kwargs.items():
            if k in docs:
                docs[k] = v
        return cls.insert(docs)

    @classmethod
    def delete(cls, oid, shard_key_value=None, **kwargs):
        """
        删除
        注意，使用该方法时，默认-1
        :return:
        :rtype:
        """
        return cls.update({"_id": oid}, {"$set": {"state": -1}})

    @classmethod
    def count(cls, sql=None, shard_key_value=None, **kwargs):
        """
        skip: 0
        limit: 0
        maxTimeMS
        :param sql:
        :param shard_key_value:
        :param kwargs:
        :return:
        """
        try:
            count = cls.dbo(shard_key_value=shard_key_value).count_documents(sql or {}, **kwargs)
        except Exception as e:
            count = cls.dbo(shard_key_value=shard_key_value).count(sql or {}, **kwargs)

        return count

    @classmethod
    def insert(cls, docs, shard_key_value=None, **kwargs):
        doc_or_docs = []
        if not docs:
            return None

        if not isinstance(docs, list):
            insert_docs = [docs]
        else:
            insert_docs = docs

        for d in insert_docs:
            d["create_at"] = utils.now()
            d["update_at"] = utils.now()
            s = vars(cls())
            s.update(d)
            doc_or_docs.append(s)

        try:
            if len(insert_docs) == 1:
                data = cls.insert_one(doc_or_docs[0], shard_key_value=None, **kwargs)
            else:
                data = cls.insert_many(doc_or_docs, shard_key_value=None, **kwargs)
        except Exception as e:
            data = cls.dbo(sql=docs, shard_key_value=shard_key_value).insert(doc_or_docs, **kwargs)

        return str(data) if not isinstance(data, list) else list(map(str, data))

    @classmethod
    def insert_one(cls, docs, shard_key_value=None, **kwargs):
        """
        ordered: True(the default) 按顺序写入，报错中断；False，无序并发写入，报错跳过
        :param docs:
        :param shard_key_value:
        :param kwargs:
        :return:
        """

        data = cls.dbo(sql=docs, shard_key_value=shard_key_value).insert_one(docs, **kwargs)
        return str(data.inserted_id)

    @classmethod
    def insert_many(cls, docs, shard_key_value=None, **kwargs):
        """
        ordered: True(the default) 按顺序写入，报错中断；False，无序并发写入，报错跳过
        :param docs:
        :param shard_key_value:
        :param kwargs:
        :return:
        """

        data = cls.dbo(sql=docs, shard_key_value=shard_key_value).insert_many(docs, **kwargs)
        data = list(map(str, data.inserted_ids))
        return data

    @classmethod
    def remove(cls, spec, shard_key_value=None, **kwargs):
        spec = cls.__check_id(spec)
        try:
            if kwargs.pop("multi", True) is True:
                deleted_count = cls.delete_many(spec, is_check_id=True, shard_key_value=shard_key_value, **kwargs)
            else:
                deleted_count = cls.delete_one(spec, is_check_id=True, shard_key_value=shard_key_value, **kwargs)
        except Exception as e:
            deleted_count = cls.dbo(sql=spec, shard_key_value=shard_key_value).remove(spec)

        return deleted_count

    @classmethod
    def delete_one(cls, spec, is_check_id=False, shard_key_value=None, **kwargs):
        if not is_check_id:
            spec = cls.__check_id(spec)
        result = cls.dbo(sql=spec, shard_key_value=shard_key_value).delete_one(spec, **kwargs)
        return result.deleted_count

    @classmethod
    def delete_many(cls, spec, is_check_id=False, shard_key_value=None, **kwargs):
        if not is_check_id:
            spec = cls.__check_id(spec)
        result = cls.dbo(sql=spec, shard_key_value=shard_key_value).delete_many(spec, **kwargs)
        return result.deleted_count

    @classmethod
    def find_one_by_oid(cls, oid, shard_key_value=None, **kwargs):
        """

        :param oid: ObjectId
        :return:
        """
        return cls.find_one({"_id": oid}, shard_key_value=shard_key_value, **kwargs)

    @classmethod
    def find_one(cls, spec=None, shard_key_value=None, **kwargs):
        """
        max_time_ms： 默认最大请求超时 60秒
        :param spec:
        :param shard_key_value:
        :param kwargs:
        :return:
        """
        spec = cls.__check_id(spec)
        kwargs.setdefault("max_time_ms", 60*1000)
        d = cls.dbo(sql=spec, shard_key_value=shard_key_value).find_one(spec or {}, **kwargs) or {}
        return cls._parse_data(cls.__parse_oid(d)) if d else {}

    @classmethod
    def find(cls, spec=None, fields=False, limit=False, skip=False, sort=False, is_cursor=False, shard_key_value=None, **kwargs):
        """
        fields: {"uid": 1}
        limit: 10
        skip: 10
        sort: {"uid": -1}
        :param spec:
        :param cursor:  True 返回游标，否则返回list
        :return:
        """
        cursor = cls.__find(spec, fields, limit, skip, sort, is_cursor, shard_key_value)
        if not is_cursor:
            data_list = [cls._parse_data(cls.__parse_oid(d), **kwargs) for d in cursor if d]
            return data_list
            # return data_list, total_page, cursor.count()
        else:
            return cursor

    @classmethod
    def distinct(cls, distinct, spec, shard_key_value=None, **kwargs):
        """
        根据spec条件，查询并返回键为distinct的列表，且不重复
        # 返回所有uid的列表，且不重复
        cls.distinct("uid", {"a": 2})
        [2, 4]

        example:
        > db.demo.find({"yyy": {"$exists": 1}})
        { "_id" : ObjectId("6138bf73dca4c58a13c81b84"), "yyy" : 400, "update_at" : 1631107579, "xxx" : 2, "a" : 2, "b" : 2 }
        { "_id" : ObjectId("6138c460dca4c58a13c81b8d"), "yyy" : 500, "update_at" : 1631107579, "xxx" : 2, "a" : 2, "b" : 3 }
        { "_id" : ObjectId("6139b44c107f8ef7d49bea5a"), "yyy" : 600, "update_at" : 1631107579, "xxx" : 4, "a" : 1, "b" : 3 }
        { "_id" : ObjectId("6139b490107f8ef7d49bea5b"), "yyy" : 400, "update_at" : 1631107579, "xxx" : 4, "a" : 2, "b" : 3 }
        > db.demo.distinct("yyy")   # 列出所有yyy的值，并且不重复
        [ 400, 500, 600 ]
        > db.demo.distinct("yyy", {"a": 2})   # 当a=2时， 列出所有yyy的值，并且不重复
        [ 400, 500 ]
        > db.demo.distinct("yyy", {"a": 2}, **{"maxTimeMS": 5000})   # 限制5s超时，防止资源耗尽
        [ 400, 500 ]
        :param spec:
        :param distinct:
        :return:
        """
        spec = cls.__check_id(spec)
        if not kwargs:  # 利用cursor游标
            result = cls.__find(spec, is_cursor=True, shard_key_value=shard_key_value).distinct(distinct)
        else:
            result = cls.dbo(shard_key_value=shard_key_value).distinct(key=distinct, filter=spec, **kwargs)

        for index, value in enumerate(result):
            if type(value) is float and int(value) == value:
                result[index] = int(value)
            elif isinstance(value, ObjectId):
                result[index] = str(value)
            else:
                result[index] = value

        return result

    @classmethod
    def aggregate(cls, pipeline, shard_key_value=None):
        return cls.dbo(shard_key_value=shard_key_value).aggregate(pipeline)

    @classmethod
    def replace_one(cls, spec, docs, shard_key_value=None, **kwargs):
        """Replace a single document matching the filter.

          >>> for doc in db.test.find({}):
          ...     print(doc)
          ...
          {u'x': 1, u'_id': ObjectId('54f4c5befba5220aa4d6dee7')}
          >>> result = db.test.replace_one({'x': 1}, {'y': 1})
          >>> result.matched_count
          1
          >>> result.modified_count
          1
          >>> for doc in db.test.find({}):
          ...     print(doc)
          ...
          {u'y': 1, u'_id': ObjectId('54f4c5befba5220aa4d6dee7')}

        The *upsert* option can be used to insert a new document if a matching
        document does not exist.

          >>> result = db.test.replace_one({'x': 1}, {'x': 1}, True)
          >>> result.matched_count
          0
          >>> result.modified_count
          0
          >>> result.upserted_id
          ObjectId('54f11e5c8891e756a6e1abd4')
          >>> db.test.find_one({'x': 1})
          {u'x': 1, u'_id': ObjectId('54f11e5c8891e756a6e1abd4')}
        """
        if docs is None:
            raise Exception('document is None!')
        if spec is None:
            raise Exception('spec is None!')

        spec = cls.__check_id(spec)
        docs.setdefault("update_at", utils.now())
        result = cls.dbo(sql=spec, shard_key_value=shard_key_value).replace_one(spec, docs, **kwargs)
        return result

    @classmethod
    def update(cls, spec=None, docs=None, check_updated_state=True, shard_key_value=None, **kwargs):
        """
        wrap collection's update
        :param spec: 更新条件
        :param docs: 更新文档
        :param check_updated_state: True 直接返回是否成功更新状态字段值(updatedExisting)
        False 返回完整的更新结果
        :param kwargs:
        """
        if docs is None:
            raise Exception('document is None!')
        if spec is None:
            raise Exception('spec is None!')

        spec = cls.__check_id(spec)

        if not any(k.startswith('$') for k in docs.keys()):
            docs = {"$set": docs}

        docs.setdefault("$set", {})['update_at'] = utils.now()

        multi = kwargs.pop("multi", False)

        try:
            if multi is True:
                result = cls.dbo(sql=spec, shard_key_value=shard_key_value).update_many(spec, docs, **kwargs)
            else:
                result = cls.dbo(sql=spec, shard_key_value=shard_key_value).update_one(spec, docs, **kwargs)
            return result.modified_count if check_updated_state else result.raw_result
        except Exception as e:
            result = cls.dbo(sql=spec, shard_key_value=shard_key_value).update(spec, docs, **kwargs)
            return result.get('nModified') if check_updated_state else result

    @classmethod
    def find_and_modify(cls, spec, docs, shard_key_value=None, **kwargs):
        spec = cls.__check_id(spec)
        docs.setdefault("$set", {})['update_at'] = utils.now()
        kwargs["new"] = kwargs.get("new", True)
        d = cls.dbo(sql=spec, shard_key_value=shard_key_value).find_and_modify(spec,
                                                                               docs,
                                                                               upsert=kwargs.pop("upsert", True),
                                                                               sort=kwargs.pop("sort", None),
                                                                               full_response=kwargs.pop("full_response", False),
                                                                               manipulate=kwargs.pop("manipulate", False), **kwargs)
        return cls._parse_data(cls.__parse_oid(d), **kwargs) if d else {}

    @classmethod
    def find_one_and_replace(cls, spec, docs, shard_key_value=None, **kwargs):
        """
        替换/覆盖/新增
        :param spec:
        :param docs:
        :param shard_key_value:
        :param kwargs:
            指定需要/不需要返回的字段
            projection：{"need_return_field": 1/True, "_id": 0/False}
            没有则新增
            upsert: True/False
            返回更新后的文档
            return_document: True/False
            按该规则排序后，更新第一条数据
            sort: {"a": -1} or [("a", -1)]
        :return:
        """
        spec = cls.__check_id(spec)
        sort = kwargs.get("sort", None)
        if sort and type(sort) == dict:
            kwargs["sort"] = list(sort.items())
        kwargs.setdefault("return_document", True)
        d = cls.dbo(sql=spec, shard_key_value=shard_key_value).find_one_and_replace(spec, docs, **kwargs)
        return cls._parse_data(cls.__parse_oid(d), **kwargs) if d else {}

    @classmethod
    def find_one_and_update(cls, spec, docs, shard_key_value=None, **kwargs):
        """
        原子更新
        :param spec:
        :param docs:
        :param shard_key_value:
        :param kwargs:
            指定需要/不需要返回的字段
            projection：{"need_return_field": 1/True, "_id": 0/False}
            没有则新增
            upsert: True/False
            返回更新后的文档
            return_document: True/False
            按该规则排序后，更新第一条数据
            sort: {"a": -1} or [("a", -1)]
        :return:
        :param spec:
        :param docs:
        :param shard_key_value:
        :param kwargs:
        :return:
        """
        spec = cls.__check_id(spec)
        sort = kwargs.get("sort", None)
        if sort and type(sort) == dict:
            kwargs["sort"] = list(sort.items())
        docs.setdefault("$set", {})['update_at'] = utils.now()
        d = cls.dbo(sql=spec, shard_key_value=shard_key_value).find_one_and_update(spec, docs, **kwargs)
        return cls._parse_data(cls.__parse_oid(d), **kwargs) if d else {}

    @classmethod
    def find_one_and_delete(cls, spec, shard_key_value=None, **kwargs):
        """
        原子删除
        :param spec:
        :param docs:
        :param shard_key_value:
        :param kwargs:
            指定需要/不需要返回的字段
            projection：{"need_return_field": 1/True, "_id": 0/False}
            返回更新后的文档
            return_document: True/False
            按该规则排序后，更新第一条数据
            sort: {"a": -1} or [("a", -1)]
        :return:
        :param spec:
        :param docs:
        :param shard_key_value:
        :param kwargs:
        :return:
        """
        spec = cls.__check_id(spec)
        sort = kwargs.get("sort", None)
        if sort and type(sort) == dict:
            kwargs["sort"] = list(sort.items())
        d = cls.dbo(sql=spec, shard_key_value=shard_key_value).find_one_and_delete(spec, **kwargs)
        return cls._parse_data(cls.__parse_oid(d), **kwargs) if d else {}

    @classmethod
    def find_list_by_page(cls, spec, page, count=10, sort={"create_at": -1}, fields=False, shard_key_value=None, **kwargs):
        """
        按照count分页，返回当前页数据和总页数
        :param spec:
        :param count:
        :param sort: must be obj like [("a", -1), ...]
        :return:
        """
        cursor = cls.__find(spec, fields=fields, is_cursor=True, shard_key_value=shard_key_value)
        data_list = [cls._parse_data(cls.__parse_oid(d), **kwargs) for d in list(cls.__meta_cursor(cursor, page, count, sort)) if d]
        total_page = cls.__get_total_page(cursor, count)
        return data_list, total_page, cursor.count()

    @classmethod
    def find_list_by_last_id(cls, spec, count=10, sort={"create_at": -1}, fields=False, last_id_field=False, shard_key_value=None, **kwargs):
        """
        list by last_id
        :param spec:
        :param count:
        :param sort: must be obj like [("a", -1), ...]
        :return:
        """
        cursor = cls.__find(spec, fields=fields, is_cursor=True, shard_key_value=shard_key_value)
        data_list = [cls._parse_data(cls.__parse_oid(d), **kwargs) for d in list(cls.__meta_cursor(cursor, count=count, sort=sort)) if d]
        last_id_field = last_id_field if last_id_field else "create_at"
        last_id = data_list[-1][last_id_field] - 1 if data_list else -1
        return data_list, last_id

    @classmethod
    def get_batch_data(cls, *vals, **kwargs):
        """
        批量获取一批数据
        :param k:
        :param vals:
        :return:
        """
        key = kwargs.pop("key", "_id")
        data_mp = {}
        skip_field_list = kwargs.get("skip_field_list", [])

        shard_key_value = kwargs.get("shard_key_value", None)
        fields = kwargs.get("fields", None)
        page = kwargs.get("page", 0)
        count = kwargs.get("count", 0)
        sort = kwargs.get("sort", {"create_at": -1})

        if vals:
            cursor = cls.__find({key: {"$in": vals}}, fields=fields, is_cursor=True, shard_key_value=shard_key_value)
            data = [cls.__parse_oid(d) for d in list(cls.__meta_cursor(cursor, 0, 0, sort)) if d]
        else:
            data = []

        for d in data:
            tmp = {}
            if skip_field_list:
                for k, v in d.items():
                    if k in skip_field_list:
                        continue
                    tmp[k] = v
            else:
                tmp = d
            data_mp[d[key]] = tmp

        return {k: cls._parse_data(v, **kwargs) for k, v in data_mp.items()}

    @classmethod
    def next_seq_id(cls, seq_name="seq_id", val=0, random_step=10):
        """Generate next value
        :param val: step value
        :return: :rtype:
        """
        if not val:
            val = random.randrange(1, random_step)

        result = cls.mongo.sequence.find_and_modify({'seq_name': seq_name}, {'$inc': {'val': val}}, True, new=True)

        if result:
            return result['val']
        return None

    @classmethod
    def __find(cls, spec=None, fields=False, limit=False, skip=False, sort=False, is_cursor=False, shard_key_value=None, **kwargs):
        """
        fields: {"uid": 1}
        limit: 10
        skip: 10
        sort: {"uid": -1}
        :param spec:
        :param cursor:  True 返回游标，否则返回list
        :return:
        """
        spec = cls.__check_id(spec)
        spec = spec or {}
        if type(sort) == dict:
            sort = list(sort.items())
        if fields and isinstance(fields, dict):
            cursor = cls.dbo(sql=spec, shard_key_value=shard_key_value).find(spec, fields)
        else:
            cursor = cls.dbo(sql=spec, shard_key_value=shard_key_value).find(spec)

        if limit is not False:
            cursor = cursor.limit(int(limit))
        if skip is not False:
            cursor = cursor.skip(int(skip))
        if sort is not False:
            cursor = cursor.sort(sort)

        if is_cursor:
            return cursor
        else:
            return list(cursor)

    @classmethod
    def _parse_data(cls, d, **kwargs):
        """
        you can rewrite this，in model
        :return:
        :rtype:
        """
        name = cls.__name__.lower().rstrip("model")
        id_name = "%s_id" % name
        if "_id" in d and id_name not in d:
            d[id_name] = str(d["_id"]) if utils.ObjectId.is_valid(d["_id"]) else d["_id"]
        return d

    @classmethod
    def __meta_cursor(cls, cursor, page=None, count=10, sort=dict()):
        sort = list(sort.items()) if type(sort) == dict else list(sort) if sort else False

        if page:
            skip = (int(page) - 1) * int(count)
            cursor.skip(skip)

        if count:
            cursor = cursor.limit(int(count))

        if sort:
            cursor = cursor.sort(sort)

        return cursor

    @classmethod
    def __get_total_page(cls, cursor, count=10):
        if not count:
            return 1
        c = cursor.count() // count
        s = cursor.count() % count
        return c + 1 if s else c

    @classmethod
    def __parse_oid(cls, d):
        if not isinstance(d, dict):
            return d

        if not d:
            d = {}
        if d and isinstance(d.get("_id"), ObjectId):
            d["_id"] = str(d["_id"])

        d = cls.__parse_date(d)
        return d

    @classmethod
    def __check_id(cls, spec):
        if not spec:
            return spec

        for k, v in spec.items():
            if k == "_id":
                if isinstance(v, dict):
                    v1 = list(v.keys())
                    if v1[0] in ["$in", "$nin"]:
                        sin = v1[0]
                        v[sin] = list(map(utils.to_ObjectId, spec[k][sin]))
                    else:
                        continue

                    spec[k] = v
                elif isinstance(v, list):
                    spec[k] = list(map(utils.to_ObjectId, spec[k]))
                else:
                    spec[k] = utils.to_ObjectId(v)
        return spec



    @classmethod
    def __parse_date(cls, d):
        if not d:
            return d

        try:
            for k in ["create_at", "update_at", "start_at", "end_at"]:
                prefix, _ = k.split("_")
                if k in d:
                    if isinstance(d[k], str) and d[k].isdigit():
                        d[k] = int(d[k])
                    d["%s_date" % prefix] = utils.timestamp2str(d[k], "%Y-%m-%d %H:%M:%S")
        except:
            pass
        return d


#!/usr/bin/env python
# -*- coding:utf-8 -*-

from bson.objectid import ObjectId
from base import BaseSchema
import time


class DemoSchema(BaseSchema):
    """
    接口名称
    """

    collection = "demo"
    is_shard = False        # 是否分表
    shard_key = "uid"       # 分表键

    def __init__(self):
        self._id = ObjectId()       # 备注信息
        self.uid = 1
        self.sex = 1

    @classmethod
    def shard_rule(cls, shard_key_value):
        """
        复写分表规则
        :param shard_key_value:
        :return:
        """
        cls.shard_key = "uid"
        return shard_key_value % 20







#!/usr/bin/env python
#encoding=utf-8

class BeanStalk:
    def __init__(self, host="localhost", port=11300, encoding='utf-8', watch='default'):
        try:
            import greenstalk
            self.conn = greenstalk.Client((host, port), encoding=encoding, watch=watch)
        except ImportError as e:
            import beanstalkc
            self.conn = beanstalkc.Connection(host, port)
            # self.conn.use(use)
            self.conn.watch(watch)
            # self.conn.ignore("default")

    def put(self, body, priority=65536, delay=0, ttr=60, tube='default'):
        """
        生产者
        :param body: job 类型为bytes或者str
        :param priority: 优先级 0-2^32 0优先级最大
        :param delay: 延迟
        :param ttr: time-to-run 超时重发
        :param tube: 向指定tube写入消息
        :return: job id
        """
        self.use(tube)
        return self.conn.put(body, priority, delay, ttr)

    def reserve(self, timeout=None):
        """
        消费者
        :param timeout: 一直处于阻塞状态 将等待多长时间来接收这个job。如果这个reserve的timeout时间到了，它将返回TimedOutError
        :return: Job类型
        """
        return self.conn.reserve(timeout)

    def delete(self, job):
        """
        删除一个job
        :param job: job或者jobId
        :return: None
        """
        self.conn.delete(job)

    def release(self, job, priority=65536, delay=0):
        """
        将一个被消费的任务重新加入队列
        :param job: 任务
        :param priority: 优先级
        :param delay: 延迟
        :return: None
        """
        self.conn.release(job, priority, delay)

    def bury(self, job, priority=65536):
        """
        把job转为buried状态
        将job放到一个特殊的FIFO队列中，之后不能被reserve命令获取，但可以用kick命令扔回工作队列中，之后就能被消费了（相当于“逻辑删除”）
        :param job: 任务
        :param priority: 优先级
        :return: None
        """
        self.conn.bury(job, priority)

    def kick(self, bound):
        """
        把一个job从buried转为ready状态
        :param bound: 一次kick的任务数，顺序从前置后
        :return: 返回受影响任务的数量
        """
        return self.conn.kick(bound)

    def kick_job(self, job):
        """
        把delayed 或者 buried 的任务转换为ready状态
        :param job: job或者id
        :return: None
        """
        self.conn.kick_job(job)

    def touch(self, job):
        """
        让任务重新计算任务超时重发ttr时间,相当于给任务延长寿命
        :param job: 任务
        :return: None
        """
        self.conn.touch(job)

    def use(self, tube='default'):
        """
        切换当前使用的tube
        每个tube都是一个独立的queue，可以使用use命令切换tube，如果切换的tube不存在，会自动创建一个
        :param tube: 管道名称 str
        :return: None
        """
        self.conn.use(tube)

    def watch(self, tube):
        """
        watch一个管道，相当于多监听一个新的tube，如果不存在，会被自动创建，可以用ignore命令取消关注tube
        当监控多个tube时，只要有一个tube有数据到来，reserve会返回
        watch和use是两个独立的动作，use一个tube不代表watching它，反之watch一个tube也不代表using它
        :param tube: 管道
        :return: 返回监听管道的数量
        """
        return self.conn.watch(tube)

    def ignore(self, tube):
        """
        取消监听指定的管道
        :param tube: 管道
        :return: 返回监听管道的数量
        """
        return self.conn.ignore(tube)

    def peek(self, jobid):
        """
        查看job的信息
        :param jobid:
        :return: 返回job类型
        """
        return self.conn.peek(jobid)

    def peek_ready(self):
        """
        获取当前管道下一个处于ready状态的任务
        :return: Job
        """
        return self.conn.peek_ready()

    def peek_delayed(self):
        """
        获取当前管道下一个处于delay状态的任务
        :return: Job
        """
        return self.conn.peek_delayed()

    def peek_buried(self):
        """
        获取当前管道最早处于buried状态的任务
        :return: Job
        """
        return self.conn.peek_buried()

    def stats_job(self, job):
        """
        查看任务的状态 job有四种状态
        ready，需要立即处理的任务，当延时 (delayed) 任务到期后会自动成为当前任务；
        delayed，延迟执行的任务, 当消费者处理任务后，可以用将消息再次放回 delayed 队列延迟执行；
        reserved，已经被消费者获取, 正在执行的任务，Beanstalkd 负责检查任务是否在 TTR(time-to-run) 内完成；
        buried，保留的任务: 任务不会被执行，也不会消失，除非有人把它 "踢" 回队列；
        :param job: job id
        :return: dict
        """
        return self.conn.stats_job(job)

    def stats_tube(self, tube):
        """
        查看管道的状态
        :param tube: 管道名称
        :return: dict
        """
        return self.conn.stats_tube(tube)

    def stats(self):
        """
        返回系统的状态
        :return: dict
        """
        return self.conn.stats()

    def tubes(self):
        """
        返回所有的管道名称
        :return list
        """
        return self.conn.tubes()

    def using(self):
        """
        返回正在使用的管道名称
        :return: str 管道名称
        """
        return self.conn.using()

    def watching(self):
        """
        返回监听的所有管道名称
        :return: list 管道名称列表
        """
        return self.conn.watching()

    def pause_tube(self, tube, delay):
        """
        暂停一个管道，delay期间的job无法使用
        :param tube: str 管道名称
        :param delay: 延迟
        :return: None
        """
        self.conn.pause_tube(tube, delay)

    def close(self):
        """
        关闭连接
        """
        self.conn.close()

if __name__ == "__main__":
    # producer
    client = BeanStalk('127.0.0.1', 11300)
    id = client.put(b'hello', delay=10)
    client.close()

    # # worker: work进程只能监听它指定的一个或多个tube通道,无论指定与否，任何work都会自动把默认的tube（default）加入监听
    # # 实现监听（watch）多个tube的方式
    # # 1：可以启动多个work进程，每个work监听七个tube
    # # 2、启动一个work，监听多个tube
    while True:
        client = BeanStalk('127.0.0.1', 11300, watch="default")
        client.watch("new")         # 新增一个监听的tube
        print(client.watching())    # 当前客户端监听的所有tube

        job = client.reserve()      # 只要一个tube有消息，消息将被获取
        print(job.id)
        print(job.body)
        client.delete(job)          # 处理完成后，收到删除消息
